// Compatibility-related imports
//import "core-js/stable";
//import "regenerator-runtime/runtime";

// Imports
import React, { useEffect, useState } from "react";
import ReactDOM from "react-dom";
import classnames from "classnames";
import { useInterval } from "react-use";
import { useSpring, animated } from "react-spring";
import { geoOrthographic, geoPath, geoDistance } from "d3-geo";
import { feature } from "topojson-client";

// types
import { Topology } from "@types/topojson-specification";

// countryData
import countryData from "./countries.json";
import markerData from "./markers.json";

interface IGlobeProps {
  size?: number;
  lat?: number;
  long?: number;
  rotation?: number;
  onClick?: () => void;
}

const markers: GeoJSON.Point[] = markerData.map((m) => ({
  type: "Point",
  coordinates: [m.coordinates[0], m.coordinates[1]],
}));

const topo: unknown = countryData;
const topology = topo as Topology;
const geojson = feature(
  topology,
  topology.objects.land
) as GeoJSON.FeatureCollection;
const countries = geojson.features;

export const Globe: React.FC<IGlobeProps> = animated(
  ({ size = 500, lat = 0, long = 0, rotation = 0, onClick }: IGlobeProps) => {
    const svgRef = React.useRef(null);

    // create geo projection to render paths
    const projection = React.useMemo(() => {
      return geoOrthographic()
        .translate([size / 2, size / 2])
        .scale(size / 2)
        .clipAngle(90)
        .rotate([rotation, 0]);
    }, [size, rotation, lat, long]);

    // return coordinates based on the projection
    const pathgen = geoPath().projection(projection);

    // find the center coordinates of the globe
    const center: [number, number] = [size / 2, size / 2];

    const colors = {
      darkBlue: "#00a9bc",
      lightBlue: "#28d9f2"
    };

    return (
      <div>
        <svg width={size} height={size} ref={svgRef} /* onClick={onClick} */>
          <circle
            cx={size / 2}
            cy={size / 2}
            r={size / 2}
            className="globe"
          />
          <g>
            {countries.map((d, i) => (
              <path key={`path-${i}`} d={pathgen(d) || ""} className="land" />
            ))}
            {markers.map((m, i) => {
              const coordinates: [number, number] = [
                m.coordinates[0],
                m.coordinates[1],
              ];
              const position = projection(coordinates);
              const x = position ? position[0] : undefined;
              const y = position ? position[1] : undefined;
              const invertedCoordinates =
                projection.invert && projection.invert(center);

              // hide the marker if rotated out of view
              const hideMarker =
                geoDistance(coordinates, invertedCoordinates || [0, 0]) > 1.4;

              return (
                <circle
                  key={`marker-${i}`}
                  className="marker"
                  cx={x}
                  cy={y}
                  r={7}
                  fillOpacity={0.4}
                  fill={hideMarker ? "none" : colors.lightBlue}
                  stroke={hideMarker ? "none" : colors.darkBlue}
                  strokeWidth={1}
                />
              );
            })}
          </g>
        </svg>
      </div>
    );
  }
);

export const RotatingGlobe: React.FC<{ size: number; duration?: number }> = ({
  size = 500,
  duration = 50000,
}) => {
  // globe positioning state
  const [orientation, setOrientation] = React.useState({
    lat: 0,
    long: 0,
    degrees: 0,
  });

  // set up spring animation
  const { lat, long, rotation } = useSpring({
    lat: orientation.lat,
    long: orientation.long,
    rotation: orientation.degrees,
    config: {
      duration,
    },
  });

  // globe click handler
  // const onClick = React.useCallback(() => {}, []);

  // kick off the initial rotation
  React.useEffect(() => {
    setOrientation({ lat: 0, long: 0, degrees: 360 });
  }, []);

  // restart the animation every {duration} milliseconds
  useInterval(() => {
    setOrientation({
      lat: Math.floor(Math.random() * size),
      long: Math.floor(Math.random() * size),
      degrees: orientation.degrees + 360,
    });
  }, duration);

  return (
    <Globe
      size={size}
      lat={lat}
      long={long}
      rotation={rotation}
      // onClick={onClick}
    />
  );
};

// Main page Vector diagram
function Diagram({className, height, width}) {
  const [_, updateState] = useState();
  const defaultXPosition = 7;
  const defaultTextLength = 60;

  useEffect(() => {
    const timeout = setTimeout(() => {
      updateState({});
    }, 100);
    return () => clearTimeout(timeout);
  }, []);

  if (!height) {
    height = "294px";
  }

  if (!width) {
    width = "900px";
  }

  return (
    <svg className={classnames(className, 'diagram')} viewBox={`0 0 683 294`} version="1.1" xmlns="http://www.w3.org/2000/svg">
      <title>Vector Diagram</title>
      <desc>A lightweight and ultra-fast tool for building observability pipelines</desc>
      <defs>
        <linearGradient x1="50%" y1="0%" x2="50%" y2="100%" id="diagram-gradient-1">
          <stop stopColor="#10E7FF" offset="0%"></stop>
          <stop stopColor="#F44AF5" offset="100%"></stop>
        </linearGradient>
        <linearGradient x1="50%" y1="0%" x2="50%" y2="100%" id="diagram-gradient-2">
          <stop stopColor="#10E7FF" offset="0%"></stop>
          <stop stopColor="#3084FF" offset="100%"></stop>
        </linearGradient>
        <linearGradient x1="50%" y1="0%" x2="50%" y2="100%" id="diagram-gradient-3">
          <stop stopColor="#10E7FF" offset="0%"></stop>
          <stop stopColor="#F44AF5" offset="100%"></stop>
        </linearGradient>
      </defs>
      <g id="Components" stroke="none" strokeWidth="1" fill="none" fillRule="evenodd">
        <g id="Group-2">
          <g id="Dividers" className="fade-in" transform="translate(192.000000, 0.000000)" fill="url(#diagram-gradient-1)" fillRule="nonzero">
            <polygon id="Line-3" points="0 0 0 293.5 2 293.5 2 0"></polygon>
            <polygon id="Line-3" points="300 0 300 293.5 302 293.5 302 0"></polygon>
          </g>
          <g id="Sink-Arrows" className="fade-in" transform="translate(466.000000, 104.000000)" fill="#000000" fillRule="nonzero">
            <path id="Line-2" d="M122.274956,56.2259684 L136.274956,63.2259684 L122.274956,70.2259684 L122.274,64.225 L120,64.2259684 L120,62.2259684 L122.274,62.225 L122.274956,56.2259684 Z M7,62.2259684 L7,64.2259684 L0,64.2259684 L0,62.2259684 L7,62.2259684 Z M17,62.2259684 L17,64.2259684 L10,64.2259684 L10,62.2259684 L17,62.2259684 Z M27,62.2259684 L27,64.2259684 L20,64.2259684 L20,62.2259684 L27,62.2259684 Z M37,62.2259684 L37,64.2259684 L30,64.2259684 L30,62.2259684 L37,62.2259684 Z M47,62.2259684 L47,64.2259684 L40,64.2259684 L40,62.2259684 L47,62.2259684 Z M57,62.2259684 L57,64.2259684 L50,64.2259684 L50,62.2259684 L57,62.2259684 Z M67,62.2259684 L67,64.2259684 L60,64.2259684 L60,62.2259684 L67,62.2259684 Z M77,62.2259684 L77,64.2259684 L70,64.2259684 L70,62.2259684 L77,62.2259684 Z M87,62.2259684 L87,64.2259684 L80,64.2259684 L80,62.2259684 L87,62.2259684 Z M97,62.2259684 L97,64.2259684 L90,64.2259684 L90,62.2259684 L97,62.2259684 Z M107,62.2259684 L107,64.2259684 L100,64.2259684 L100,62.2259684 L107,62.2259684 Z M117,62.2259684 L117,64.2259684 L110,64.2259684 L110,62.2259684 L117,62.2259684 Z"></path>
            <path id="Line-2" d="M39,116.225968 L53,123.225968 L39,130.225968 L39,116.225968 Z M7,122.225968 L7,124.225968 L0,124.225968 L0,122.225968 L7,122.225968 Z M17,122.225968 L17,124.225968 L10,124.225968 L10,122.225968 L17,122.225968 Z M27,122.225968 L27,124.225968 L20,124.225968 L20,122.225968 L27,122.225968 Z M37,122.225968 L37,124.225968 L30,124.225968 L30,122.225968 L37,122.225968 Z"></path>
            <path id="Line-2" d="M39,-6.5 L53,0.5 L39,7.5 L39,-6.5 Z M14,-0.5 L14,1.5 L0,1.5 L0,-0.5 L14,-0.5 Z M38,-0.5 L38,1.5 L24,1.5 L24,-0.5 L38,-0.5 Z"></path>
          </g>
          <g id="Source-Arrows" className="fade-in" transform="translate(86.500000, 101.638384)" fill="#000000" fillRule="nonzero">
            <path id="Line-4" d="M133.155668,-6.64242161 L147.221371,0.224600628 L133.288331,7.35694982 L133.231,1.357 L79.0095209,1.87104747 L78.9905689,-0.128862732 L133.212,-0.642 L133.155668,-6.64242161 Z"></path>
            <path id="Line-5" d="M133.326364,-7.03026802 L147,0.587584848 L132.701536,6.95578182 L132.965891,1.03234055 C108.504239,1.65538762 90.3526539,11.2783109 78.3993935,29.9305268 L78.3993935,29.9305268 L69.7273462,43.4895749 L69.6187869,43.6451529 L69.5458669,43.724737 L69.4168931,43.8314804 L69.3240426,43.815818 L67.1087866,47.0978826 L63.3331136,52.764216 C52.470948,69.1254278 31.2128344,73.4282409 -0.23475072,65.833672 L-0.23475072,65.833672 L-1.20680632,65.5989213 L-0.737304886,63.6548101 L0.23475072,63.8895608 C30.9549174,71.3084586 51.3634704,67.1776035 61.6668864,51.6580122 C64.6807471,47.1183638 66.6584223,44.1678077 67.7614023,42.5579921 L67.7614023,42.5579921 L68.607827,41.3458454 C68.6566703,41.2787491 68.6992179,41.2215209 68.7363717,41.1730296 L68.7363717,41.1730296 L68.8539133,41.0277005 L68.995713,40.888649 L69.03146,40.861551 L77.0927923,28.268647 C89.4433616,9.38586562 108.133644,-0.370070055 133.054629,-0.970462358 L133.326364,-7.03026802 Z"></path>
            <path id="Line-6" d="M133.099945,-6.86262491 L147,0.333795697 L132.90283,7.13598737 L132.986229,1.19576402 C112.502595,2.15785305 96.5785037,8.52317621 85.1610844,20.2781533 C71.2435841,34.6071267 63.6740952,56.9688315 62.4992556,87.4001939 L62.4992556,87.4001939 L62.4606781,88.3994495 L60.4621669,88.3222945 L60.5007444,87.3230389 C61.6930285,56.4398177 69.419375,33.6147256 83.7264219,18.8846887 C95.544694,6.71700753 111.991765,0.157433767 133.014439,-0.807695954 L133.099945,-6.86262491 Z"></path>
          </g>
          <g id="Sinks" transform="translate(527.000000, 5.000000)">
            <a href="/docs/reference/configuration/sinks/prometheus_exporter">
              <g id="Sink-3" transform="translate(0.000000, 184.000000)">
                <circle id="Oval-Copy" fill="#000000" cx="37.5" cy="37.5" r="37.5"></circle>
                <path d="M16.2259484,20 C18.123312,20 19.5799758,20.3335638 20.5959834,21.0007013 C21.6119911,21.6678388 22.1199872,22.7603367 22.1199872,24.2782276 C22.1199872,25.8083596 21.6058706,26.9130983 20.5776219,27.5924769 C19.5493732,28.2718555 18.0804686,28.6115397 16.1708639,28.6115397 L15.2711508,28.6115397 L15.2711508,32.8714058 L12.406758,32.8714058 L12.406758,20.3305069 C13.0310519,20.2080963 13.692059,20.1224102 14.3897992,20.073446 C15.0875394,20.0244817 15.699583,20 16.2259484,20 Z M16.4095633,22.4420784 C16.2014653,22.4420784 15.9964307,22.4481989 15.7944533,22.4604399 C15.5924759,22.472681 15.4180435,22.4849218 15.2711508,22.4971629 L15.2711508,26.1694613 L16.1708639,26.1694613 C17.1623894,26.1694613 17.9090826,26.0348117 18.4109659,25.7655084 C18.9128492,25.4962052 19.1637871,24.9943295 19.1637871,24.2598661 C19.1637871,23.9048755 19.0995225,23.6110946 18.9709914,23.3785145 C18.8424603,23.1459344 18.6588472,22.9592611 18.4201466,22.818489 C18.181446,22.6777169 17.8907253,22.5797899 17.5479758,22.5247051 C17.2052262,22.4696204 16.8257592,22.4420784 16.4095633,22.4420784 Z M29.8685368,25.636978 C29.6237157,25.5757727 29.3360552,25.5115081 29.0055467,25.4441823 C28.6750382,25.3768565 28.3200529,25.3431941 27.9405802,25.3431941 C27.7692054,25.3431941 27.5641708,25.3584952 27.3254702,25.3890979 C27.0867696,25.4197005 26.9062167,25.4533629 26.7838062,25.4900861 L26.7838062,32.8714058 L24.0479439,32.8714058 L24.0479439,23.7273828 C24.5375861,23.5560081 25.1159673,23.3938165 25.7831049,23.2408033 C26.4502424,23.0877901 27.1938754,23.0112847 28.0140261,23.0112847 C28.1609188,23.0112847 28.3384115,23.0204653 28.5465094,23.0388269 C28.7546074,23.0571885 28.9627022,23.0816702 29.1708001,23.1122729 C29.3788981,23.1428755 29.5869929,23.1795981 29.7950909,23.2224418 C30.0031888,23.2652855 30.1806814,23.3173092 30.3275741,23.3785145 L29.8685368,25.636978 Z M40.6467325,28.0239719 C40.6467325,28.7829174 40.5365647,29.4775869 40.3162257,30.1080013 C40.0958867,30.7384156 39.777624,31.277014 39.3614281,31.7238126 C38.9452322,32.1706111 38.4464167,32.5164157 37.8649665,32.7612369 C37.2835164,33.006058 36.6316899,33.1284667 35.9094676,33.1284667 C35.1994864,33.1284667 34.5537804,33.006058 33.9723303,32.7612369 C33.3908801,32.5164157 32.8920646,32.1706111 32.4758687,31.7238126 C32.0596728,31.277014 31.7352897,30.7384156 31.5027096,30.1080013 C31.2701295,29.4775869 31.1538412,28.7829174 31.1538412,28.0239719 C31.1538412,27.2650265 31.2731898,26.5734172 31.5118903,25.9491234 C31.7505909,25.3248295 32.0810945,24.7923516 32.5034109,24.3516736 C32.9257273,23.9109956 33.4276031,23.5682511 34.0090532,23.32343 C34.5905034,23.0786089 35.2239685,22.9562002 35.9094676,22.9562002 C36.6072078,22.9562002 37.2467934,23.0786089 37.8282435,23.32343 C38.4096937,23.5682511 38.9085092,23.9109956 39.3247051,24.3516736 C39.740901,24.7923516 40.0652841,25.3248295 40.2978642,25.9491234 C40.5304443,26.5734172 40.6467325,27.2650265 40.6467325,28.0239719 Z M37.8557858,28.0239719 C37.8557858,27.1793391 37.6874738,26.5152718 37.3508448,26.0317501 C37.0142157,25.5482284 36.5337615,25.3064712 35.9094676,25.3064712 C35.2851738,25.3064712 34.8016593,25.5482284 34.4589098,26.0317501 C34.1161602,26.5152718 33.944788,27.1793391 33.944788,28.0239719 C33.944788,28.8686048 34.1161602,29.5387925 34.4589098,30.0345553 C34.8016593,30.5303181 35.2851738,30.7781957 35.9094676,30.7781957 C36.5337615,30.7781957 37.0142157,30.5303181 37.3508448,30.0345553 C37.6874738,29.5387925 37.8557858,28.8686048 37.8557858,28.0239719 Z M48.0647753,27.7485496 C48.0647753,26.8671936 47.9515472,26.2429091 47.7250877,25.8756774 C47.4986281,25.5084457 47.1099804,25.3248326 46.5591329,25.3248326 C46.3877581,25.3248326 46.2102655,25.3340133 46.0266497,25.3523749 C45.8430338,25.3707365 45.6655412,25.392158 45.4941664,25.4166401 L45.4941664,32.8714058 L42.7583041,32.8714058 L42.7583041,23.5437679 C42.9908842,23.4825627 43.2632436,23.4182981 43.5753905,23.3509723 C43.8875374,23.2836465 44.218041,23.2224421 44.5669111,23.1673573 C44.9157812,23.1122726 45.2738267,23.0694295 45.6410583,23.0388269 C46.00829,23.0082243 46.3693957,22.9929232 46.7243864,22.9929232 C47.4221265,22.9929232 47.9882669,23.0816695 48.4228244,23.2591648 C48.8573818,23.4366601 49.2154273,23.6478152 49.4969716,23.8926363 C49.8886854,23.611092 50.3385375,23.3907563 50.8465413,23.2316226 C51.3545451,23.0724888 51.8227585,22.9929232 52.2511954,22.9929232 C53.0223819,22.9929232 53.6558471,23.1000308 54.1516098,23.3142493 C54.6473726,23.5284678 55.0421407,23.8314293 55.335926,24.2231431 C55.6297114,24.6148569 55.8316858,25.0800101 55.9418553,25.6186165 C56.0520248,26.157223 56.1071087,26.7570257 56.1071087,27.4180427 L56.1071087,32.8714058 L53.3712464,32.8714058 L53.3712464,27.7485496 C53.3712464,26.8671936 53.2580183,26.2429091 53.0315588,25.8756774 C52.8050993,25.5084457 52.4164516,25.3248326 51.8656041,25.3248326 C51.7187114,25.3248326 51.5136768,25.3615553 51.2504941,25.4350016 C50.9873114,25.5084479 50.7700359,25.6002545 50.5986611,25.710424 C50.6843485,25.9919683 50.7394325,26.2888094 50.7639146,26.6009563 C50.7883967,26.9131033 50.8006376,27.246667 50.8006376,27.6016576 L50.8006376,32.8714058 L48.0647753,32.8714058 L48.0647753,27.7485496 Z M57.759643,26.1510998 L63.0844756,26.1510998 L63.0844756,28.6299012 L57.759643,28.6299012 L57.759643,26.1510998 Z M8,46.9873127 C8,46.1304388 8.13158938,45.3806854 8.39477208,44.73803 C8.65795477,44.0953745 9.00375941,43.5598364 9.43219637,43.1313994 C9.86063332,42.7029625 10.3533284,42.3785794 10.9102965,42.1582404 C11.4672645,41.9379014 12.0395253,41.8277335 12.627096,41.8277335 C13.9980942,41.8277335 15.0814114,42.2469834 15.87708,43.0854957 C16.6727486,43.924008 17.070577,45.1572759 17.070577,46.7853363 C17.070577,46.94447 17.0644565,47.1189025 17.0522155,47.3086388 C17.0399744,47.4983752 17.0277336,47.6666872 17.0154925,47.8135799 L10.8093083,47.8135799 C10.8705135,48.3766684 11.1336923,48.8234603 11.5988524,49.1539688 C12.0640125,49.4844773 12.688297,49.649729 13.4717246,49.649729 C13.9736079,49.649729 14.466303,49.6038258 14.9498247,49.5120179 C15.4333464,49.4202099 15.8281145,49.3069819 16.1341409,49.1723303 L16.5013707,51.3940708 C16.3544781,51.4675171 16.1586241,51.5409623 15.913803,51.6144087 C15.6689819,51.687855 15.3966225,51.7521196 15.0967166,51.8072043 C14.7968107,51.8622891 14.4754878,51.9081924 14.1327383,51.9449155 C13.7899887,51.9816387 13.4472443,52 13.1044947,52 C12.2353798,52 11.4795059,51.8714708 10.8368505,51.6144087 C10.1941951,51.3573465 9.66171713,51.0054214 9.2394007,50.5586229 C8.81708428,50.1118244 8.50494204,49.5824066 8.30296462,48.9703538 C8.1009872,48.3583011 8,47.697294 8,46.9873127 Z M14.4265222,45.9407077 C14.4142811,45.7081276 14.3744983,45.4816715 14.3071725,45.2613325 C14.2398467,45.0409935 14.1357992,44.8451395 13.9950271,44.6737647 C13.854255,44.50239 13.6767623,44.3616199 13.4625438,44.2514504 C13.2483254,44.1412809 12.9820864,44.086197 12.6638189,44.086197 C12.3577925,44.086197 12.0946138,44.1382207 11.8742748,44.2422697 C11.6539358,44.3463187 11.4703227,44.4840285 11.32343,44.6554033 C11.1765374,44.826778 11.0633093,45.0256922 10.9837424,45.2521517 C10.9041756,45.4786113 10.8460314,45.7081276 10.8093083,45.9407077 L14.4265222,45.9407077 Z M19.0903411,39.679439 L21.8262034,39.2387631 L21.8262034,42.0847944 L25.1129104,42.0847944 L25.1129104,44.3616194 L21.8262034,44.3616194 L21.8262034,47.7584954 C21.8262034,48.333825 21.9271906,48.7928577 22.129168,49.1356073 C22.3311454,49.4783568 22.7381544,49.649729 23.3502072,49.649729 C23.6439925,49.649729 23.9469541,49.6221871 24.2591011,49.5671023 C24.571248,49.5120176 24.8558483,49.4355121 25.1129104,49.3375837 L25.4985018,51.4675167 C25.1679932,51.6021683 24.8007671,51.7184566 24.3968122,51.8163851 C23.9928574,51.9143135 23.4971021,51.963277 22.9095314,51.963277 C22.162827,51.963277 21.544663,51.8622898 21.0550207,51.6603124 C20.5653785,51.458335 20.1736706,51.1767949 19.8798852,50.8156838 C19.5860999,50.4545726 19.3810653,50.0169614 19.2647753,49.5028371 C19.1484852,48.9887128 19.0903411,48.4195091 19.0903411,47.7952184 L19.0903411,39.679439 Z M27.3346509,51.7429391 L27.3346509,37.9350972 L30.0705132,37.4944214 L30.0705132,42.1215174 C30.2541291,42.0603121 30.4897659,42.002168 30.7774307,41.9470832 C31.0650955,41.8919984 31.3435753,41.8644565 31.6128785,41.8644565 C32.3963061,41.8644565 33.0481326,41.9715641 33.5683774,42.1857826 C34.0886223,42.4000011 34.504812,42.7029627 34.8169589,43.0946764 C35.1291058,43.4863902 35.3494415,43.9515434 35.4779726,44.4901498 C35.6065037,45.0287563 35.6707682,45.628559 35.6707682,46.289576 L35.6707682,51.7429391 L32.934906,51.7429391 L32.934906,46.6200829 C32.934906,45.7387269 32.8216779,45.1144424 32.5952184,44.7472107 C32.3687588,44.379979 31.949509,44.196366 31.3374562,44.196366 C31.0926351,44.196366 30.8631187,44.2177875 30.6489002,44.2606312 C30.4346817,44.3034749 30.241888,44.3493781 30.0705132,44.3983424 L30.0705132,51.7429391 L27.3346509,51.7429391 Z M37.6905324,46.9873127 C37.6905324,46.1304388 37.8221217,45.3806854 38.0853044,44.73803 C38.3484871,44.0953745 38.6942918,43.5598364 39.1227287,43.1313994 C39.5511657,42.7029625 40.0438608,42.3785794 40.6008288,42.1582404 C41.1577969,41.9379014 41.7300576,41.8277335 42.3176283,41.8277335 C43.6886266,41.8277335 44.7719437,42.2469834 45.5676124,43.0854957 C46.363281,43.924008 46.7611093,45.1572759 46.7611093,46.7853363 C46.7611093,46.94447 46.7549889,47.1189025 46.7427478,47.3086388 C46.7305068,47.4983752 46.7182659,47.6666872 46.7060249,47.8135799 L40.4998406,47.8135799 C40.5610459,48.3766684 40.8242246,48.8234603 41.2893848,49.1539688 C41.7545449,49.4844773 42.3788294,49.649729 43.1622569,49.649729 C43.6641402,49.649729 44.1568353,49.6038258 44.640357,49.5120179 C45.1238787,49.4202099 45.5186469,49.3069819 45.8246733,49.1723303 L46.1919031,51.3940708 C46.0450104,51.4675171 45.8491565,51.5409623 45.6043354,51.6144087 C45.3595142,51.687855 45.0871548,51.7521196 44.787249,51.8072043 C44.4873431,51.8622891 44.1660202,51.9081924 43.8232706,51.9449155 C43.4805211,51.9816387 43.1377767,52 42.7950271,52 C41.9259121,52 41.1700383,51.8714708 40.5273828,51.6144087 C39.8847274,51.3573465 39.3522495,51.0054214 38.9299331,50.5586229 C38.5076166,50.1118244 38.1954744,49.5824066 37.993497,48.9703538 C37.7915196,48.3583011 37.6905324,47.697294 37.6905324,46.9873127 Z M44.1170545,45.9407077 C44.1048135,45.7081276 44.0650306,45.4816715 43.9977048,45.2613325 C43.930379,45.0409935 43.8263316,44.8451395 43.6855595,44.6737647 C43.5447873,44.50239 43.3672947,44.3616199 43.1530762,44.2514504 C42.9388577,44.1412809 42.6726187,44.086197 42.3543513,44.086197 C42.0483249,44.086197 41.7851461,44.1382207 41.5648071,44.2422697 C41.3444681,44.3463187 41.1608551,44.4840285 41.0139624,44.6554033 C40.8670697,44.826778 40.7538416,45.0256922 40.6742748,45.2521517 C40.5947079,45.4786113 40.5365638,45.7081276 40.4998406,45.9407077 L44.1170545,45.9407077 Z M57.1169908,51.4124323 C56.6518306,51.5470839 56.0520279,51.6725528 55.3175646,51.7888428 C54.5831012,51.9051329 53.8119263,51.963277 53.0040166,51.963277 C52.1838658,51.963277 51.5014372,51.8531092 50.9567102,51.6327702 C50.4119833,51.4124312 49.9804925,51.1033491 49.6622251,50.7055148 C49.3439576,50.3076805 49.1175015,49.8333467 48.9828499,49.2824992 C48.8481982,48.7316517 48.7808734,48.1257285 48.7808734,47.4647115 L48.7808734,42.0847944 L51.5167357,42.0847944 L51.5167357,47.1342047 C51.5167357,48.0155607 51.633024,48.652086 51.8656041,49.0437998 C52.0981841,49.4355136 52.5327351,49.6313675 53.16927,49.6313675 C53.3651269,49.6313675 53.5732217,49.6221869 53.7935607,49.6038253 C54.0138997,49.5854637 54.2097537,49.5640422 54.3811285,49.5395601 L54.3811285,42.0847944 L57.1169908,42.0847944 L57.1169908,51.4124323 Z M62.4601849,49.7782595 C62.9620682,49.7782595 63.3170535,49.729296 63.5251514,49.6313675 C63.7332494,49.5334391 63.8372968,49.3437056 63.8372968,49.0621613 C63.8372968,48.8418223 63.7026472,48.6490286 63.433344,48.4837743 C63.1640407,48.3185201 62.7539715,48.1318468 62.203124,47.9237488 C61.7746871,47.7646151 61.3860394,47.5993633 61.0371693,47.4279885 C60.6882992,47.2566137 60.391458,47.0515791 60.1466369,46.8128785 C59.9018158,46.574178 59.7120823,46.2895777 59.5774307,45.9590692 C59.4427791,45.6285607 59.3754543,45.2307323 59.3754543,44.7655722 C59.3754543,43.8597341 59.7120782,43.1436431 60.3853363,42.6172777 C61.0585944,42.0909123 61.9827802,41.8277335 63.1579216,41.8277335 C63.7454923,41.8277335 64.3085724,41.8797572 64.8471788,41.9838062 C65.3857853,42.0878552 65.8142158,42.2010832 66.1324833,42.3234938 L65.6550845,44.4534268 C65.336817,44.3432573 64.9910124,44.2453304 64.6176602,44.159643 C64.244308,44.0739556 63.8250581,44.0311125 63.359898,44.0311125 C62.5030241,44.0311125 62.0745936,44.2698095 62.0745936,44.7472107 C62.0745936,44.8573802 62.0929549,44.9553072 62.129678,45.0409946 C62.1664012,45.126682 62.2398464,45.2093079 62.3500159,45.2888747 C62.4601854,45.3684416 62.6101361,45.4541277 62.7998725,45.5459356 C62.9896089,45.6377435 63.2313661,45.7387307 63.5251514,45.8489002 C64.1249632,46.0692392 64.6207185,46.2865147 65.0124323,46.5007332 C65.404146,46.7149517 65.7132281,46.9475282 65.9396876,47.1984699 C66.1661471,47.4494115 66.3252785,47.7278914 66.4170864,48.0339178 C66.5088943,48.3399442 66.5547976,48.6949294 66.5547976,49.0988843 C66.5547976,50.0536866 66.1967521,50.7758981 65.4806503,51.2655403 C64.7645485,51.7551826 63.7516164,52 62.4418234,52 C61.5849495,52 60.8719187,51.9265548 60.3027096,51.7796621 C59.7335005,51.6327694 59.3387324,51.5103607 59.1183934,51.4124323 L59.5774307,49.1906917 C60.0425908,49.3743076 60.5199848,49.5181378 61.009627,49.6221868 C61.4992693,49.7262358 61.9827837,49.7782595 62.4601849,49.7782595 Z" id="Prom-etheus" fill="#FFFFFF" fillRule="nonzero"></path>
              </g>
            </a>
            <a href="/docs/reference/configuration/sinks/elasticsearch">
              <g id="Sink-2" transform="translate(81.000000, 124.000000)">
                <circle id="Oval-Copy" fill="#000000" cx="37.5" cy="37.5" r="37.5"></circle>
                <path d="M11,33.8509704 L11,21.4814827 L19.3534203,21.4814827 L19.3534203,23.8197264 L13.7844734,23.8197264 L13.7844734,26.247216 L18.7286987,26.247216 L18.7286987,28.5319122 L13.7844734,28.5319122 L13.7844734,31.5127267 L19.7639516,31.5127267 L19.7639516,33.8509704 L11,33.8509704 Z M25.3864461,34.0294623 C24.6129774,34.0175628 23.9852872,33.9342674 23.5033567,33.7795737 C23.0214262,33.6248799 22.6406473,33.407717 22.3610086,33.1280783 C22.0813699,32.8484396 21.8909805,32.5093084 21.7898346,32.1106745 C21.6886886,31.7120406 21.6381164,31.2628406 21.6381164,30.7630608 L21.6381164,20.4283805 L24.2976456,20 L24.2976456,30.2275851 C24.2976456,30.4655755 24.3154946,30.6797636 24.3511931,30.8701559 C24.3868917,31.0605482 24.4553129,31.2211893 24.5564588,31.352084 C24.6576047,31.4829787 24.8033716,31.5900728 24.9937639,31.6733694 C25.1841562,31.756666 25.439992,31.8102131 25.761279,31.8340121 L25.3864461,34.0294623 Z M31.1338848,31.9946548 C31.3956742,31.9946548 31.6455604,31.9887051 31.8835507,31.9768056 C32.1215411,31.9649061 32.3119306,31.9470571 32.4547248,31.923258 L32.4547248,29.9062997 C32.3476291,29.8825007 32.186988,29.858702 31.9727967,29.834903 C31.7586054,29.8111039 31.5622662,29.7992046 31.3837735,29.7992046 C31.1338836,29.7992046 30.8988716,29.8140788 30.6787305,29.8438276 C30.4585894,29.8735763 30.2652251,29.9300982 30.0986319,30.0133948 C29.9320386,30.0966915 29.8011459,30.2097352 29.7059497,30.3525294 C29.6107536,30.4953237 29.5631562,30.6738138 29.5631562,30.8880051 C29.5631562,31.3044882 29.7029735,31.5930472 29.9826122,31.7536907 C30.2622508,31.9143342 30.6460046,31.9946548 31.1338848,31.9946548 Z M30.9196946,24.2124085 C31.7050628,24.2124085 32.3595265,24.3016536 32.8831053,24.4801464 C33.4066841,24.6586391 33.8261359,24.9144749 34.1414731,25.2476615 C34.4568104,25.580848 34.679923,25.9854256 34.8108177,26.4614063 C34.9417124,26.937387 35.0071588,27.4669077 35.0071588,28.0499841 L35.0071588,33.5832326 C34.6263742,33.6665292 34.0968535,33.7646988 33.418581,33.8777442 C32.7403084,33.9907896 31.9192539,34.0473115 30.9553929,34.0473115 C30.3485175,34.0473115 29.798173,33.9937645 29.304343,33.8866688 C28.810513,33.7795731 28.3851116,33.6040579 28.028126,33.3601177 C27.6711404,33.1161776 27.3974556,32.7978702 27.2070633,32.4051861 C27.016671,32.012502 26.9214763,31.5305787 26.9214763,30.9594018 C26.9214763,30.412024 27.0315452,29.9479497 27.2516863,29.5671651 C27.4718274,29.1863805 27.766336,28.8829474 28.1352211,28.6568565 C28.5041062,28.4307657 28.9265328,28.2671497 29.4025135,28.1660038 C29.8784943,28.0648579 30.3723169,28.0142857 30.8839962,28.0142857 C31.2290822,28.0142857 31.5354902,28.0291599 31.8032294,28.0589087 C32.0709686,28.0886575 32.2881315,28.1273303 32.4547248,28.1749284 L32.4547248,27.9250398 C32.4547248,27.4728581 32.3178824,27.1099282 32.0441934,26.8362393 C31.7705045,26.5625503 31.2945309,26.4257079 30.6162584,26.4257079 C30.1640766,26.4257079 29.7178514,26.4584311 29.2775692,26.5238785 C28.837287,26.5893258 28.4565081,26.6815457 28.1352211,26.8005409 L27.7960865,24.6586382 C27.9507803,24.6110402 28.1441446,24.560468 28.3761852,24.5069201 C28.6082258,24.4533723 28.8610868,24.4057749 29.1347757,24.3641266 C29.4084646,24.3224783 29.6970236,24.2867803 30.0004613,24.2570315 C30.3038991,24.2272827 30.6103071,24.2124085 30.9196946,24.2124085 Z M40.1120267,31.9411072 C40.599907,31.9411072 40.9449879,31.8935099 41.1472797,31.7983137 C41.3495715,31.7031176 41.4507159,31.5186778 41.4507159,31.2449889 C41.4507159,31.0307975 41.3198231,30.8433829 41.0580337,30.6827394 C40.7962443,30.5220959 40.3976164,30.340631 39.8621381,30.1383392 C39.4456549,29.9836454 39.0678509,29.8230043 38.7287146,29.6564111 C38.3895783,29.4898178 38.1010193,29.2905039 37.863029,29.0584633 C37.6250386,28.8264226 37.4405988,28.549763 37.3097041,28.228476 C37.1788094,27.907189 37.113363,27.5204604 37.113363,27.0682787 C37.113363,26.1877143 37.4405949,25.4916029 38.0950684,24.9799236 C38.7495419,24.4682443 39.6479421,24.2124085 40.7902959,24.2124085 C41.3614728,24.2124085 41.9088424,24.2629807 42.4324213,24.3641266 C42.9560001,24.4652725 43.372477,24.5753414 43.6818645,24.6943366 L43.2177856,26.7648425 C42.9083981,26.6577468 42.5722417,26.5625521 42.2093064,26.4792555 C41.8463711,26.3959589 41.4388187,26.3543112 40.986637,26.3543112 C40.1536707,26.3543112 39.7371938,26.5863483 39.7371938,27.0504295 C39.7371938,27.1575252 39.7550428,27.2527199 39.7907413,27.3360165 C39.8264399,27.4193132 39.8978359,27.4996337 40.0049316,27.5769806 C40.1120273,27.6543275 40.2577942,27.7376228 40.4422367,27.8268692 C40.6266793,27.9161156 40.8616912,28.0142852 41.1472797,28.1213808 C41.7303561,28.3355722 42.2122794,28.5467855 42.593064,28.755027 C42.9738485,28.9632686 43.2743069,29.1893561 43.494448,29.4332962 C43.7145891,29.6772363 43.8692805,29.9479463 43.9585269,30.2454343 C44.0477733,30.5429223 44.0923958,30.8880031 44.0923958,31.2806872 C44.0923958,32.2088497 43.7443401,32.9109108 43.0482183,33.3868915 C42.3520964,33.8628722 41.367426,34.1008591 40.0941775,34.1008591 C39.2612112,34.1008591 38.5680747,34.029463 38.0147471,33.8866688 C37.4614194,33.7438746 37.0776657,33.6248812 36.8634744,33.529685 L37.3097041,31.3699332 C37.7618858,31.548426 38.2259601,31.6882432 38.7019408,31.7893891 C39.1779216,31.890535 39.6479455,31.9411072 40.1120267,31.9411072 Z M45.9844098,22.1240535 L48.6439389,21.6956729 L48.6439389,24.4622972 L51.8389437,24.4622972 L51.8389437,26.6755966 L48.6439389,26.6755966 L48.6439389,29.9776965 C48.6439389,30.5369738 48.7421085,30.9831991 48.9384505,31.3163856 C49.1347926,31.6495721 49.5304456,31.8161629 50.1254216,31.8161629 C50.41101,31.8161629 50.7055187,31.7893894 51.0089564,31.7358416 C51.3123941,31.6822937 51.5890538,31.6079228 51.8389437,31.5127267 L52.2137766,33.5832326 C51.8924896,33.7141273 51.5355094,33.827171 51.1428253,33.9223672 C50.7501412,34.0175633 50.2682179,34.0651607 49.697041,34.0651607 C48.9711704,34.0651607 48.3702537,33.9669911 47.894273,33.7706491 C47.4182922,33.574307 47.0375134,33.3006222 46.7519249,32.9495864 C46.4663365,32.5985506 46.2670225,32.1731492 46.1539771,31.6733694 C46.0409317,31.1735896 45.9844098,30.6202673 45.9844098,30.0133948 L45.9844098,22.1240535 Z M56.6582246,33.8509704 L53.9986955,33.8509704 L53.9986955,24.4622972 L56.6582246,24.4622972 L56.6582246,33.8509704 Z M56.9081133,21.7492205 C56.9081133,22.2371008 56.750447,22.6208545 56.4351098,22.9004932 C56.1197725,23.1801318 55.7479181,23.3199491 55.3195355,23.3199491 C54.8911528,23.3199491 54.5192984,23.1801318 54.2039612,22.9004932 C53.8886239,22.6208545 53.7309577,22.2371008 53.7309577,21.7492205 C53.7309577,21.2613402 53.8886239,20.8775865 54.2039612,20.5979478 C54.5192984,20.3183091 54.8911528,20.1784919 55.3195355,20.1784919 C55.7479181,20.1784919 56.1197725,20.3183091 56.4351098,20.5979478 C56.750447,20.8775865 56.9081133,21.2613402 56.9081133,21.7492205 Z M58.7108813,29.1566338 C58.7108813,28.4783612 58.8209502,27.8387717 59.0410913,27.237846 C59.2612324,26.6369203 59.5795398,26.1133494 59.9960229,25.6671174 C60.4125061,25.2208855 60.918228,24.8668801 61.5132039,24.6050907 C62.1081799,24.3433013 62.7864423,24.2124085 63.5480115,24.2124085 C64.0477912,24.2124085 64.5059158,24.2570311 64.922399,24.3462774 C65.3388821,24.4355238 65.7434597,24.5634417 66.1361438,24.730035 L65.582819,26.8540885 C65.3329291,26.7588923 65.0592442,26.6755969 64.7617563,26.6041998 C64.4642683,26.5328027 64.1310868,26.4971047 63.7622017,26.4971047 C62.9768335,26.4971047 62.390791,26.7410411 62.0040566,27.2289214 C61.6173223,27.7168017 61.423958,28.359366 61.423958,29.1566338 C61.423958,30.0014996 61.6054229,30.6559633 61.9683583,31.1200445 C62.3312936,31.5841258 62.9649334,31.8161629 63.8692969,31.8161629 C64.1905839,31.8161629 64.5356647,31.7864146 64.9045498,31.726917 C65.2734349,31.6674194 65.6125661,31.5722246 65.9219535,31.4413299 L66.2967865,33.618931 C65.987399,33.7498257 65.6006705,33.8628694 65.1365892,33.9580655 C64.672508,34.0532617 64.1608364,34.1008591 63.601559,34.1008591 C62.7447937,34.1008591 62.0070346,33.9729411 61.3882596,33.7171015 C60.7694847,33.4612618 60.2607879,33.1132061 59.862154,32.672924 C59.4635201,32.2326418 59.1719863,31.7120456 58.9875437,31.1111199 C58.8031012,30.5101943 58.7108813,29.8587054 58.7108813,29.1566338 Z M14.2753261,50.8402482 C14.7632064,50.8402482 15.1082872,50.7926508 15.3105791,50.6974547 C15.5128709,50.6022585 15.6140153,50.4178187 15.6140153,50.1441298 C15.6140153,49.9299385 15.4831225,49.7425239 15.2213331,49.5818804 C14.9595437,49.4212369 14.5609158,49.2397719 14.0254375,49.0374801 C13.6089543,48.8827864 13.2311503,48.7221453 12.892014,48.555552 C12.5528777,48.3889588 12.2643187,48.1896448 12.0263283,47.9576042 C11.788338,47.7255636 11.6038982,47.4489039 11.4730035,47.1276169 C11.3421088,46.8063299 11.2766624,46.4196014 11.2766624,45.9674197 C11.2766624,45.0868553 11.6038943,44.3907439 12.2583678,43.8790646 C12.9128413,43.3673853 13.8112415,43.1115495 14.9535953,43.1115495 C15.5247722,43.1115495 16.0721418,43.1621217 16.5957206,43.2632676 C17.1192995,43.3644135 17.5357764,43.4744824 17.8451639,43.5934776 L17.381085,45.6639835 C17.0716975,45.5568878 16.7355411,45.4616931 16.3726058,45.3783964 C16.0096705,45.2950998 15.6021181,45.2534521 15.1499364,45.2534521 C14.3169701,45.2534521 13.9004932,45.4854892 13.9004932,45.9495705 C13.9004932,46.0566661 13.9183422,46.1518609 13.9540407,46.2351575 C13.9897393,46.3184541 14.0611353,46.3987747 14.168231,46.4761215 C14.2753267,46.5534684 14.4210936,46.6367638 14.6055361,46.7260102 C14.7899787,46.8152566 15.0249906,46.9134261 15.3105791,47.0205218 C15.8936555,47.2347131 16.3755788,47.4459264 16.7563633,47.654168 C17.1371479,47.8624096 17.4376063,48.088497 17.6577474,48.3324372 C17.8778885,48.5763773 18.0325799,48.8470873 18.1218263,49.1445752 C18.2110727,49.4420632 18.2556952,49.7871441 18.2556952,50.1798282 C18.2556952,51.1079906 17.9076395,51.8100517 17.2115177,52.2860325 C16.5153958,52.7620132 15.5307254,53 14.2574769,53 C13.4245106,53 12.7313741,52.928604 12.1780465,52.7858097 C11.6247188,52.6430155 11.2409651,52.5240221 11.0267738,52.428826 L11.4730035,50.2690741 C11.9251852,50.4475669 12.3892595,50.5873842 12.8652402,50.6885301 C13.341221,50.789676 13.8112449,50.8402482 14.2753261,50.8402482 Z M19.7907254,48.1271715 C19.7907254,47.2942052 19.9186433,46.5653706 20.174483,45.9406459 C20.4303226,45.3159212 20.766479,44.795325 21.1829621,44.3788419 C21.5994453,43.9623587 22.0783937,43.6470262 22.6198218,43.4328349 C23.1612499,43.2186435 23.7175441,43.1115495 24.288721,43.1115495 C25.6214671,43.1115495 26.6745587,43.5191019 27.4480274,44.3342189 C28.2214961,45.1493359 28.6082246,46.3481944 28.6082246,47.9308304 C28.6082246,48.0855242 28.602275,48.2550898 28.5903754,48.4395323 C28.5784759,48.6239748 28.5665766,48.7875908 28.5546771,48.930385 L22.5216513,48.930385 C22.5811489,49.4777628 22.8369847,49.9120888 23.2891664,50.2333758 C23.7413481,50.5546628 24.3482145,50.7153038 25.1097836,50.7153038 C25.5976639,50.7153038 26.0766123,50.6706813 26.5466433,50.5814349 C27.0166743,50.4921885 27.400428,50.3821196 27.697916,50.2512249 L28.0548998,52.4109768 C27.9121056,52.4823739 27.7217161,52.5537699 27.4837257,52.625167 C27.2457354,52.6965641 26.980975,52.7590357 26.6894368,52.8125835 C26.3978986,52.8661314 26.085541,52.9107539 25.7523544,52.9464524 C25.4191679,52.982151 25.0859864,53 24.7527999,53 C23.9079341,53 23.1731498,52.8750569 22.5484251,52.625167 C21.9237003,52.3752771 21.406079,52.0331711 20.9955457,51.5988387 C20.5850123,51.1645063 20.2815791,50.6498598 20.085237,50.0548839 C19.888895,49.4599079 19.7907254,48.8173436 19.7907254,48.1271715 Z M26.0379415,47.1097677 C26.0260419,46.8836769 25.9873691,46.6635391 25.9219217,46.4493478 C25.8564744,46.2351564 25.75533,46.044767 25.6184855,45.8781737 C25.4816411,45.7115805 25.3091006,45.574738 25.1008591,45.4676424 C24.8926175,45.3605467 24.6338068,45.3069997 24.3244193,45.3069997 C24.0269314,45.3069997 23.7710956,45.3575719 23.5569042,45.4587178 C23.3427129,45.5598637 23.1642228,45.6937313 23.0214286,45.8603245 C22.8786343,46.0269178 22.7685655,46.2202821 22.6912186,46.4404232 C22.6138717,46.6605643 22.5573498,46.8836769 22.5216513,47.1097677 L26.0379415,47.1097677 Z M34.2485682,50.8937957 C34.5103577,50.8937957 34.7602438,50.8878461 34.9982342,50.8759465 C35.2362245,50.864047 35.426614,50.846198 35.5694082,50.822399 L35.5694082,48.8054407 C35.4623125,48.7816416 35.3016714,48.7578429 35.0874801,48.7340439 C34.8732888,48.7102449 34.6769497,48.6983455 34.4984569,48.6983455 C34.248567,48.6983455 34.013555,48.7132197 33.7934139,48.7429685 C33.5732728,48.7727173 33.3799086,48.8292392 33.2133153,48.9125358 C33.046722,48.9958324 32.9158293,49.1088762 32.8206332,49.2516704 C32.725437,49.3944646 32.6778396,49.5729547 32.6778396,49.787146 C32.6778396,50.2036292 32.8176569,50.4921882 33.0972956,50.6528317 C33.3769343,50.8134752 33.760688,50.8937957 34.2485682,50.8937957 Z M34.034378,43.1115495 C34.8197462,43.1115495 35.4742099,43.2007945 35.9977887,43.3792873 C36.5213676,43.5577801 36.9408193,43.8136159 37.2561565,44.1468024 C37.5714938,44.4799889 37.7946064,44.8845665 37.9255011,45.3605472 C38.0563958,45.836528 38.1218422,46.3660486 38.1218422,46.949125 L38.1218422,52.4823735 C37.7410576,52.5656702 37.211537,52.6638397 36.5332644,52.7768851 C35.8549918,52.8899306 35.0339374,52.9464524 34.0700764,52.9464524 C33.4632009,52.9464524 32.9128564,52.8929054 32.4190264,52.7858097 C31.9251964,52.6787141 31.499795,52.5031988 31.1428094,52.2592587 C30.7858239,52.0153185 30.512139,51.6970112 30.3217467,51.3043271 C30.1313544,50.911643 30.0361597,50.4297197 30.0361597,49.8585428 C30.0361597,49.3111649 30.1462286,48.8470907 30.3663697,48.4663061 C30.5865108,48.0855215 30.8810195,47.7820883 31.2499045,47.5559975 C31.6187896,47.3299066 32.0412162,47.1662907 32.5171969,47.0651448 C32.9931777,46.9639989 33.4870003,46.9134267 33.9986796,46.9134267 C34.3437656,46.9134267 34.6501737,46.9283008 34.9179128,46.9580496 C35.185652,46.9877984 35.4028149,47.0264713 35.5694082,47.0740694 L35.5694082,46.8241807 C35.5694082,46.371999 35.4325658,46.0090691 35.1588769,45.7353802 C34.8851879,45.4616913 34.4092143,45.3248489 33.7309418,45.3248489 C33.2787601,45.3248489 32.8325348,45.3575721 32.3922526,45.4230194 C31.9519704,45.4884668 31.5711916,45.5806866 31.2499045,45.6996818 L30.91077,43.5577792 C31.0654637,43.5101811 31.258828,43.4596089 31.4908686,43.4060611 C31.7229092,43.3525133 31.9757702,43.3049159 32.2494591,43.2632676 C32.523148,43.2216193 32.811707,43.1859212 33.1151448,43.1561724 C33.4185825,43.1264237 33.7249905,43.1115495 34.034378,43.1115495 Z M46.1896755,45.717531 C45.9516851,45.6580334 45.6720506,45.5955619 45.3507636,45.5301145 C45.0294766,45.4646672 44.6843957,45.431944 44.3155107,45.431944 C44.1489174,45.431944 43.9496035,45.4468182 43.7175628,45.476567 C43.4855222,45.5063158 43.310007,45.539039 43.1910118,45.5747375 L43.1910118,52.7501114 L40.5314827,52.7501114 L40.5314827,43.8612154 C41.0074634,43.6946221 41.5697072,43.5369559 42.218231,43.3882119 C42.8667548,43.2394679 43.5896397,43.165097 44.3869074,43.165097 C44.5297016,43.165097 44.7022421,43.1740215 44.9045339,43.1918708 C45.1068257,43.2097201 45.3091145,43.2335188 45.5114063,43.2632676 C45.7136981,43.2930164 45.9159869,43.3287144 46.1182787,43.3703627 C46.3205705,43.412011 46.493111,43.4625832 46.6359052,43.5220808 L46.1896755,45.717531 Z M47.4391187,48.0557747 C47.4391187,47.3775022 47.5491876,46.7379126 47.7693287,46.136987 C47.9894698,45.5360613 48.3077771,45.0124903 48.7242603,44.5662584 C49.1407434,44.1200264 49.6464654,43.766021 50.2414413,43.5042316 C50.8364172,43.2424422 51.5146796,43.1115495 52.2762488,43.1115495 C52.7760286,43.1115495 53.2341532,43.156172 53.6506363,43.2454184 C54.0671195,43.3346648 54.4716971,43.4625827 54.8643812,43.6291759 L54.3110563,45.7532294 C54.0611664,45.6580332 53.7874816,45.5747379 53.4899936,45.5033408 C53.1925057,45.4319436 52.8593241,45.3962456 52.4904391,45.3962456 C51.7050708,45.3962456 51.1190283,45.6401821 50.732294,46.1280624 C50.3455596,46.6159426 50.1521954,47.258507 50.1521954,48.0557747 C50.1521954,48.9006406 50.3336603,49.5551043 50.6965956,50.0191855 C51.0595309,50.4832667 51.6931708,50.7153038 52.5975342,50.7153038 C52.9188212,50.7153038 53.2639021,50.6855555 53.6327871,50.6260579 C54.0016722,50.5665603 54.3408034,50.4713656 54.6501909,50.3404709 L55.0250239,52.5180719 C54.7156364,52.6489666 54.3289078,52.7620103 53.8648266,52.8572065 C53.4007454,52.9524026 52.8890737,53 52.3297964,53 C51.473031,53 50.7352719,52.8720821 50.116497,52.6162424 C49.497722,52.3604028 48.9890252,52.0123471 48.5903913,51.5720649 C48.1917575,51.1317827 47.9002236,50.6111866 47.7157811,50.0102609 C47.5313386,49.4093352 47.4391187,48.7578463 47.4391187,48.0557747 Z M56.8099427,52.7501114 L56.8099427,39.3275215 L59.4694718,38.8991409 L59.4694718,43.3971365 C59.6479646,43.3376389 59.8770269,43.281117 60.1566656,43.2275692 C60.4363043,43.1740214 60.7070143,43.1472479 60.9688037,43.1472479 C61.7303729,43.1472479 62.3640127,43.2513671 62.8697423,43.4596087 C63.3754718,43.6678502 63.7800494,43.9623589 64.0834871,44.3431435 C64.3869248,44.7239281 64.601113,45.176103 64.7260579,45.6996818 C64.8510029,46.2232607 64.9134744,46.8063283 64.9134744,47.4489023 L64.9134744,52.7501114 L62.2539453,52.7501114 L62.2539453,47.7701877 C62.2539453,46.9134224 62.1438764,46.306556 61.9237353,45.9495705 C61.7035942,45.5925849 61.2960418,45.4140948 60.7010659,45.4140948 C60.4630755,45.4140948 60.2399629,45.4349187 60.0317213,45.476567 C59.8234797,45.5182153 59.6360651,45.5628378 59.4694718,45.6104359 L59.4694718,52.7501114 L56.8099427,52.7501114 Z" id="Elasticsearch" fill="#FFFFFF" fillRule="nonzero"></path>
              </g>
            </a>
            <a href="/docs/reference/configuration/sinks/aws_s3">
              <g id="Sink-1" transform="translate(1.000000, 58.000000)">
                <circle id="Oval-Copy-2" fill="#000000" cx="37.5" cy="37.5" r="37.5"></circle>
                <path d="M25.8375,45.7875 C26.6625041,45.7875 27.3437473,45.7187507 27.88125,45.58125 C28.4187527,45.4437493 28.8499984,45.2562512 29.175,45.01875 C29.5000016,44.7812488 29.7249994,44.5000016 29.85,44.175 C29.9750006,43.8499984 30.0375,43.487502 30.0375,43.0875 C30.0375,42.2374957 29.6375,41.5312528 28.8375,40.96875 C28.037496,40.4062472 26.6625097,39.8000033 24.7125,39.15 C23.8624957,38.8499985 23.0125042,38.5062519 22.1625,38.11875 C21.3124958,37.7312481 20.5500034,37.2437529 19.875,36.65625 C19.1999966,36.0687471 18.6500021,35.3562542 18.225,34.51875 C17.7999979,33.6812458 17.5875,32.6625 17.5875,31.4625 C17.5875,30.262494 17.8124977,29.1812548 18.2625,28.21875 C18.7125022,27.2562452 19.3499959,26.4375034 20.175,25.7625 C21.0000041,25.0874966 21.9999941,24.5687518 23.175,24.20625 C24.3500059,23.8437482 25.6749926,23.6625 27.15,23.6625 C28.9000087,23.6625 30.4124936,23.8499981 31.6875,24.225 C32.9625064,24.6000019 34.0124959,25.0124978 34.8375,25.4625 L33.15,30.075 C32.4249964,29.6999981 31.6187544,29.3687514 30.73125,29.08125 C29.8437456,28.7937486 28.7750063,28.65 27.525,28.65 C26.124993,28.65 25.1187531,28.8437481 24.50625,29.23125 C23.8937469,29.6187519 23.5875,30.212496 23.5875,31.0125 C23.5875,31.4875024 23.6999989,31.8874984 23.925,32.2125 C24.1500011,32.5375016 24.4687479,32.8312487 24.88125,33.09375 C25.2937521,33.3562513 25.7687473,33.5937489 26.30625,33.80625 C26.8437527,34.0187511 27.4374967,34.2374989 28.0875,34.4625 C29.4375068,34.9625025 30.612495,35.4562476 31.6125,35.94375 C32.612505,36.4312524 33.4437467,36.9999968 34.10625,37.65 C34.7687533,38.3000033 35.2624984,39.0624956 35.5875,39.9375 C35.9125016,40.8125044 36.075,41.8749938 36.075,43.125 C36.075,45.5500121 35.2250085,47.4312433 33.525,48.76875 C31.8249915,50.1062567 29.2625171,50.775 25.8375,50.775 C24.6874942,50.775 23.6500046,50.7062507 22.725,50.56875 C21.7999954,50.4312493 20.9812536,50.262501 20.26875,50.0625 C19.5562464,49.862499 18.9437526,49.6500011 18.43125,49.425 C17.9187474,49.1999989 17.4875017,48.987501 17.1375,48.7875 L18.7875,44.1375 C19.5625039,44.5625021 20.5187443,44.9437483 21.65625,45.28125 C22.7937557,45.6187517 24.1874917,45.7875 25.8375,45.7875 Z M46.125,50.8125 C45.4499966,50.8125 44.7375038,50.7687504 43.9875,50.68125 C43.2374962,50.5937496 42.5125035,50.4750008 41.8125,50.325 C41.1124965,50.1749993 40.4750029,50.0125009 39.9,49.8375 C39.3249971,49.6625 38.8750016,49.5000007 38.55,49.35 L39.6375,44.7 C40.2875032,44.9750014 41.1187449,45.2687484 42.13125,45.58125 C43.1437551,45.8937516 44.3999925,46.05 45.9,46.05 C47.6250086,46.05 48.887496,45.7250033 49.6875,45.075 C50.487504,44.4249967 50.8875,43.5500055 50.8875,42.45 C50.8875,41.7749966 50.7437514,41.2062523 50.45625,40.74375 C50.1687486,40.2812477 49.7750025,39.9062514 49.275,39.61875 C48.7749975,39.3312486 48.1812534,39.1312506 47.49375,39.01875 C46.8062466,38.9062494 46.0750039,38.85 45.3,38.85 L43.125,38.85 L43.125,34.35 L45.6,34.35 C46.1500027,34.35 46.6812474,34.3000005 47.19375,34.2 C47.7062526,34.0999995 48.162498,33.9312512 48.5625,33.69375 C48.962502,33.4562488 49.2812488,33.1312521 49.51875,32.71875 C49.7562512,32.3062479 49.875,31.7875031 49.875,31.1625 C49.875,30.6874976 49.775001,30.2750017 49.575,29.925 C49.374999,29.5749983 49.1187516,29.2875011 48.80625,29.0625 C48.4937484,28.8374989 48.1312521,28.6687506 47.71875,28.55625 C47.3062479,28.4437494 46.8875021,28.3875 46.4625,28.3875 C45.3874946,28.3875 44.3937546,28.5499984 43.48125,28.875 C42.5687454,29.2000016 41.7375038,29.5999976 40.9875,30.075 L39,25.9875 C39.400002,25.7374988 39.8687473,25.4750014 40.40625,25.2 C40.9437527,24.9249986 41.5374968,24.6750011 42.1875,24.45 C42.8375032,24.2249989 43.5312463,24.0375007 44.26875,23.8875 C45.0062537,23.7374992 45.7874959,23.6625 46.6125,23.6625 C48.1375076,23.6625 49.4562444,23.8437482 50.56875,24.20625 C51.6812556,24.5687518 52.5999964,25.0812467 53.325,25.74375 C54.05,26.4062533 54.5874982,27.1812456 54.9375,28.06875 C55.2875017,28.9562544 55.4625,29.9249948 55.4625,30.975 C55.4625,32.0000051 55.1750029,32.9937452 54.6,33.95625 C54.0249971,34.9187548 53.2500049,35.6499975 52.275,36.15 C53.6250067,36.7000028 54.6687463,37.5187446 55.40625,38.60625 C56.1437537,39.6937554 56.5125,40.9999924 56.5125,42.525 C56.5125,43.725006 56.312502,44.8312449 55.9125,45.84375 C55.512498,46.8562551 54.8875042,47.7312463 54.0375,48.46875 C53.1874957,49.2062537 52.1062566,49.7812479 50.79375,50.19375 C49.4812434,50.6062521 47.925009,50.8125 46.125,50.8125 Z" id="S3" fill="#FFFFFF" fillRule="nonzero"></path>
              </g>
            </a>
            <a href="/docs/reference/configuration/sinks">
              <path className="fade-in" d="M17.154,15.382 C17.6380024,15.382 18.0376651,15.3416671 18.353,15.261 C18.6683349,15.1803329 18.9213324,15.070334 19.112,14.931 C19.3026676,14.791666 19.4346663,14.6266676 19.508,14.436 C19.5813337,14.2453324 19.618,14.0326678 19.618,13.798 C19.618,13.2993308 19.3833357,12.8850016 18.914,12.555 C18.4446643,12.2249983 17.6380057,11.8693352 16.494,11.488 C15.9953308,11.3119991 15.4966692,11.1103345 14.998,10.883 C14.4993308,10.6556655 14.052002,10.3696684 13.656,10.025 C13.259998,9.68033161 12.9373346,9.26233579 12.688,8.771 C12.4386654,8.27966421 12.314,7.68200352 12.314,6.978 C12.314,6.27399648 12.4459987,5.63966949 12.71,5.075 C12.9740013,4.51033051 13.3479976,4.03000198 13.832,3.634 C14.3160024,3.23799802 14.9026632,2.93366773 15.592,2.721 C16.2813368,2.50833227 17.0586623,2.402 17.924,2.402 C18.9506718,2.402 19.8379963,2.5119989 20.586,2.732 C21.3340037,2.9520011 21.9499976,3.19399868 22.434,3.458 L21.444,6.164 C21.0186645,5.9439989 20.5456693,5.74966751 20.025,5.581 C19.5043307,5.41233249 18.877337,5.328 18.144,5.328 C17.3226626,5.328 16.7323351,5.44166553 16.373,5.669 C16.0136649,5.89633447 15.834,6.24466432 15.834,6.714 C15.834,6.99266806 15.8999993,7.22733238 16.032,7.418 C16.1640007,7.60866762 16.3509988,7.78099923 16.593,7.935 C16.8350012,8.08900077 17.1136651,8.22833271 17.429,8.353 C17.7443349,8.47766729 18.0926648,8.60599934 18.474,8.738 C19.266004,9.0313348 19.9553304,9.32099857 20.542,9.607 C21.1286696,9.89300143 21.6163314,10.2266648 22.005,10.608 C22.3936686,10.9893352 22.6833324,11.4366641 22.874,11.95 C23.0646676,12.4633359 23.16,13.086663 23.16,13.82 C23.16,15.2426738 22.6613383,16.3463294 21.664,17.131 C20.6666617,17.9156706 19.1633434,18.308 17.154,18.308 C16.47933,18.308 15.8706694,18.2676671 15.328,18.187 C14.7853306,18.1063329 14.3050021,18.0073339 13.887,17.89 C13.4689979,17.7726661 13.1096682,17.6480007 12.809,17.516 C12.5083318,17.3839993 12.2553344,17.2593339 12.05,17.142 L13.018,14.414 C13.4726689,14.6633346 14.0336633,14.886999 14.701,15.085 C15.3683367,15.283001 16.1859952,15.382 17.154,15.382 Z M28.858,18 L25.58,18 L25.58,6.428 L28.858,6.428 L28.858,18 Z M29.166,3.084 C29.166,3.68533634 28.9716686,4.15833161 28.583,4.503 C28.1943314,4.84766839 27.7360026,5.02 27.208,5.02 C26.6799974,5.02 26.2216686,4.84766839 25.833,4.503 C25.4443314,4.15833161 25.25,3.68533634 25.25,3.084 C25.25,2.48266366 25.4443314,2.00966839 25.833,1.665 C26.2216686,1.32033161 26.6799974,1.148 27.208,1.148 C27.7360026,1.148 28.1943314,1.32033161 28.583,1.665 C28.9716686,2.00966839 29.166,2.48266366 29.166,3.084 Z M31.938,6.824 C32.4953361,6.66266586 33.2139956,6.51233403 34.094,6.373 C34.9740044,6.23366597 35.8979952,6.164 36.866,6.164 C37.8486716,6.164 38.6663301,6.29233205 39.319,6.549 C39.9716699,6.80566795 40.4886648,7.16866432 40.87,7.638 C41.2513352,8.10733568 41.5226659,8.66466344 41.684,9.31 C41.8453341,9.95533656 41.926,10.673996 41.926,11.466 L41.926,18 L38.648,18 L38.648,11.862 C38.648,10.8059947 38.5086681,10.0580022 38.23,9.618 C37.9513319,9.1779978 37.4306705,8.958 36.668,8.958 C36.4333322,8.958 36.1840013,8.96899989 35.92,8.991 C35.6559987,9.01300011 35.4213344,9.03866652 35.216,9.068 L35.216,18 L31.938,18 L31.938,6.824 Z M48.174,10.608 C48.4966683,10.2559982 48.8303316,9.88933524 49.175,9.508 C49.5196684,9.12666476 49.8533317,8.74900187 50.176,8.375 C50.4986683,8.00099813 50.8029986,7.64533502 51.089,7.308 C51.3750014,6.97066498 51.6206656,6.67733458 51.826,6.428 L55.72,6.428 C54.9426628,7.32267114 54.1836704,8.17699593 53.443,8.991 C52.7023296,9.80500407 51.8920044,10.6446623 51.012,11.51 C51.4520022,11.906002 51.9066643,12.3789972 52.376,12.929 C52.8453357,13.4790027 53.2999978,14.0473304 53.74,14.634 C54.1800022,15.2206696 54.5833315,15.8073304 54.95,16.394 C55.3166685,16.9806696 55.6246654,17.5159976 55.874,18 L52.112,18 C51.8773322,17.6186648 51.6096682,17.1970023 51.309,16.735 C51.0083318,16.2729977 50.689335,15.8110023 50.352,15.349 C50.014665,14.8869977 49.6590019,14.4433355 49.285,14.018 C48.9109981,13.5926645 48.5406685,13.2333348 48.174,12.94 L48.174,18 L44.896,18 L44.896,1.456 L48.174,0.928 L48.174,10.608 Z M60.956,15.646 C61.5573363,15.646 61.9826654,15.5873339 62.232,15.47 C62.4813346,15.3526661 62.606,15.125335 62.606,14.788 C62.606,14.5239987 62.4446683,14.293001 62.122,14.095 C61.7993317,13.896999 61.3080033,13.6733346 60.648,13.424 C60.1346641,13.2333324 59.6690021,13.0353344 59.251,12.83 C58.8329979,12.6246656 58.4773348,12.3790014 58.184,12.093 C57.8906652,11.8069986 57.6633341,11.466002 57.502,11.07 C57.3406659,10.673998 57.26,10.1973361 57.26,9.64 C57.26,8.55466124 57.6633293,7.69666982 58.47,7.066 C59.2766707,6.43533018 60.383993,6.12 61.792,6.12 C62.4960035,6.12 63.1706634,6.18233271 63.816,6.307 C64.4613366,6.43166729 64.9746648,6.5673326 65.356,6.714 L64.784,9.266 C64.4026648,9.13399934 63.9883356,9.01666718 63.541,8.914 C63.0936644,8.81133282 62.5913361,8.76 62.034,8.76 C61.0073282,8.76 60.494,9.04599714 60.494,9.618 C60.494,9.75000066 60.5159998,9.86733282 60.56,9.97 C60.6040002,10.0726672 60.6919993,10.1716662 60.824,10.267 C60.9560007,10.3623338 61.1356655,10.4649994 61.363,10.575 C61.5903345,10.6850005 61.8799982,10.8059993 62.232,10.938 C62.9506703,11.2020013 63.5446643,11.462332 64.014,11.719 C64.4833357,11.9756679 64.8536653,12.2543318 65.125,12.555 C65.3963347,12.8556682 65.5869995,13.1893315 65.697,13.556 C65.8070006,13.9226685 65.862,14.3479976 65.862,14.832 C65.862,15.9760057 65.4330043,16.8413304 64.575,17.428 C63.7169957,18.0146696 62.5033412,18.308 60.934,18.308 C59.9073282,18.308 59.0530034,18.2200009 58.371,18.044 C57.6889966,17.8679991 57.2160013,17.7213339 56.952,17.604 L57.502,14.942 C58.0593361,15.1620011 58.6313304,15.3343327 59.218,15.459 C59.8046696,15.5836673 60.3839971,15.646 60.956,15.646 Z" fill="#000000" fillRule="nonzero"></path>
            </a>
          </g>
          <g id="Sources" transform="translate(0.000000, 7.000000)">
            <a href="/docs/reference/configuration/sources/statsd">
              <g id="Source-3" transform="translate(82.000000, 182.000000)">
                <circle id="Oval-Copy-2" fill="#000000" cx="37.5" cy="37.5" r="37.5"></circle>
                <path d="M10.683,43.501 C11.1450023,43.501 11.5264985,43.4625004 11.8275,43.3855 C12.1285015,43.3084996 12.3699991,43.2035007 12.552,43.0705 C12.7340009,42.9374993 12.8599996,42.7800009 12.93,42.598 C13.0000004,42.4159991 13.035,42.2130011 13.035,41.989 C13.035,41.5129976 12.811,41.1175016 12.363,40.8025 C11.9149978,40.4874984 11.1450055,40.1480018 10.053,39.784 C9.57699762,39.6159992 9.10100238,39.4235011 8.625,39.2065 C8.14899762,38.9894989 7.72200189,38.7165016 7.344,38.3875 C6.96599811,38.0584984 6.65800119,37.6595023 6.42,37.1905 C6.18199881,36.7214977 6.063,36.1510034 6.063,35.479 C6.063,34.8069966 6.18899874,34.2015027 6.441,33.6625 C6.69300126,33.1234973 7.04999769,32.6650019 7.512,32.287 C7.97400231,31.9089981 8.53399671,31.618501 9.192,31.4155 C9.85000329,31.212499 10.5919959,31.111 11.418,31.111 C12.3980049,31.111 13.2449964,31.2159989 13.959,31.426 C14.6730036,31.6360011 15.2609977,31.8669987 15.723,32.119 L14.778,34.702 C14.371998,34.4919989 13.9205025,34.3065008 13.4235,34.1455 C12.9264975,33.9844992 12.3280035,33.904 11.628,33.904 C10.8439961,33.904 10.2805017,34.0124989 9.9375,34.2295 C9.59449829,34.4465011 9.423,34.7789978 9.423,35.227 C9.423,35.4930013 9.48599937,35.7169991 9.612,35.899 C9.73800063,36.0810009 9.91649884,36.2454993 10.1475,36.3925 C10.3785012,36.5395007 10.6444985,36.6724994 10.9455,36.7915 C11.2465015,36.9105006 11.5789982,37.0329994 11.943,37.159 C12.6990038,37.4390014 13.3569972,37.7154986 13.917,37.9885 C14.4770028,38.2615014 14.9424981,38.5799982 15.3135,38.944 C15.6845019,39.3080018 15.9609991,39.7349976 16.143,40.225 C16.3250009,40.7150025 16.416,41.3099965 16.416,42.01 C16.416,43.3680068 15.9400048,44.4214963 14.988,45.1705 C14.0359952,45.9195037 12.6010096,46.294 10.683,46.294 C10.0389968,46.294 9.45800259,46.2555004 8.94,46.1785 C8.42199741,46.1014996 7.96350199,46.0070006 7.5645,45.895 C7.165498,45.7829994 6.82250143,45.6640006 6.5355,45.538 C6.24849856,45.4119994 6.00700098,45.2930006 5.811,45.181 L6.735,42.577 C7.16900217,42.8150012 7.70449681,43.0284991 8.3415,43.2175 C8.97850318,43.4065009 9.75899538,43.501 10.683,43.501 Z M18.621,32.203 L21.75,31.699 L21.75,34.954 L25.509,34.954 L25.509,37.558 L21.75,37.558 L21.75,41.443 C21.75,42.1010033 21.8654988,42.625998 22.0965,43.018 C22.3275012,43.410002 22.7929965,43.606 23.493,43.606 C23.8290017,43.606 24.1754982,43.5745003 24.5325,43.5115 C24.8895018,43.4484997 25.2149985,43.3610006 25.509,43.249 L25.95,45.685 C25.5719981,45.8390008 25.1520023,45.9719994 24.69,46.084 C24.2279977,46.1960006 23.6610034,46.252 22.989,46.252 C22.1349957,46.252 21.4280028,46.1365012 20.868,45.9055 C20.3079972,45.6744988 19.8600017,45.3525021 19.524,44.9395 C19.1879983,44.5264979 18.9535007,44.0260029 18.8205,43.438 C18.6874993,42.8499971 18.621,42.199 18.621,41.485 L18.621,32.203 Z M31.851,43.816 C32.1590015,43.816 32.4529986,43.8090001 32.733,43.795 C33.0130014,43.7809999 33.2369992,43.7600001 33.405,43.732 L33.405,41.359 C33.2789994,41.3309999 33.0900013,41.3030001 32.838,41.275 C32.5859987,41.2469999 32.355001,41.233 32.145,41.233 C31.8509985,41.233 31.5745013,41.2504998 31.3155,41.2855 C31.0564987,41.3205002 30.829001,41.3869995 30.633,41.485 C30.436999,41.5830005 30.2830006,41.7159992 30.171,41.884 C30.0589994,42.0520008 30.003,42.2619987 30.003,42.514 C30.003,43.0040025 30.1674984,43.3434991 30.4965,43.5325 C30.8255016,43.7215009 31.2769971,43.816 31.851,43.816 Z M31.599,34.66 C32.5230046,34.66 33.2929969,34.7649989 33.909,34.975 C34.5250031,35.185001 35.0184981,35.485998 35.3895,35.878 C35.7605019,36.270002 36.0229992,36.7459972 36.177,37.306 C36.3310008,37.8660028 36.408,38.4889966 36.408,39.175 L36.408,45.685 C35.9599978,45.7830005 35.337004,45.8984993 34.539,46.0315 C33.740996,46.1645007 32.7750057,46.231 31.641,46.231 C30.9269964,46.231 30.2795029,46.1680006 29.6985,46.042 C29.1174971,45.9159994 28.6170021,45.7095014 28.197,45.4225 C27.7769979,45.1354986 27.4550011,44.7610023 27.231,44.299 C27.0069989,43.8369977 26.895,43.2700034 26.895,42.598 C26.895,41.9539968 27.0244987,41.4080022 27.2835,40.96 C27.5425013,40.5119978 27.8889978,40.1550013 28.323,39.889 C28.7570022,39.6229987 29.2539972,39.4305006 29.814,39.3115 C30.3740028,39.1924994 30.954997,39.133 31.557,39.133 C31.963002,39.133 32.3234984,39.1504998 32.6385,39.1855 C32.9535016,39.2205002 33.208999,39.2659997 33.405,39.322 L33.405,39.028 C33.405,38.4959973 33.2440016,38.0690016 32.922,37.747 C32.5999984,37.4249984 32.040004,37.264 31.242,37.264 C30.7099973,37.264 30.1850026,37.3024996 29.667,37.3795 C29.1489974,37.4565004 28.7010019,37.5649993 28.323,37.705 L27.924,35.185 C28.1060009,35.1289997 28.3334986,35.0695003 28.6065,35.0065 C28.8795014,34.9434997 29.1769984,34.8875002 29.499,34.8385 C29.8210016,34.7894998 30.1604982,34.7475002 30.5175,34.7125 C30.8745018,34.6774998 31.2349982,34.66 31.599,34.66 Z M39.138,32.203 L42.267,31.699 L42.267,34.954 L46.026,34.954 L46.026,37.558 L42.267,37.558 L42.267,41.443 C42.267,42.1010033 42.3824988,42.625998 42.6135,43.018 C42.8445012,43.410002 43.3099965,43.606 44.01,43.606 C44.3460017,43.606 44.6924982,43.5745003 45.0495,43.5115 C45.4065018,43.4484997 45.7319985,43.3610006 46.026,43.249 L46.467,45.685 C46.0889981,45.8390008 45.6690023,45.9719994 45.207,46.084 C44.7449977,46.1960006 44.1780034,46.252 43.506,46.252 C42.6519957,46.252 41.9450028,46.1365012 41.385,45.9055 C40.8249972,45.6744988 40.3770017,45.3525021 40.041,44.9395 C39.7049983,44.5264979 39.4705007,44.0260029 39.3375,43.438 C39.2044993,42.8499971 39.138,42.199 39.138,41.485 L39.138,32.203 Z M51.738,43.753 C52.3120029,43.753 52.7179988,43.6970006 52.956,43.585 C53.1940012,43.4729994 53.313,43.2560016 53.313,42.934 C53.313,42.6819987 53.1590015,42.4615009 52.851,42.2725 C52.5429985,42.0834991 52.0740031,41.8700012 51.444,41.632 C50.9539975,41.4499991 50.509502,41.261001 50.1105,41.065 C49.711498,40.868999 49.3720014,40.6345014 49.092,40.3615 C48.8119986,40.0884986 48.5950008,39.7630019 48.441,39.385 C48.2869992,39.0069981 48.21,38.5520027 48.21,38.02 C48.21,36.9839948 48.5949961,36.165003 49.365,35.563 C50.1350038,34.960997 51.1919933,34.66 52.536,34.66 C53.2080034,34.66 53.8519969,34.7194994 54.468,34.8385 C55.0840031,34.9575006 55.5739982,35.0869993 55.938,35.227 L55.392,37.663 C55.0279982,37.5369994 54.6325021,37.4250005 54.2055,37.327 C53.7784979,37.2289995 53.2990027,37.18 52.767,37.18 C51.7869951,37.18 51.297,37.4529973 51.297,37.999 C51.297,38.1250006 51.3179998,38.2369995 51.36,38.335 C51.4020002,38.4330005 51.4859994,38.5274995 51.612,38.6185 C51.7380006,38.7095005 51.9094989,38.8074995 52.1265,38.9125 C52.3435011,39.0175005 52.6199983,39.1329994 52.956,39.259 C53.6420034,39.5110013 54.2089978,39.7594988 54.657,40.0045 C55.1050022,40.2495012 55.4584987,40.5154986 55.7175,40.8025 C55.9765013,41.0895014 56.1584995,41.4079982 56.2635,41.758 C56.3685005,42.1080017 56.421,42.5139977 56.421,42.976 C56.421,44.0680055 56.0115041,44.8939972 55.1925,45.454 C54.3734959,46.0140028 53.2150075,46.294 51.717,46.294 C50.7369951,46.294 49.9215033,46.2100008 49.2705,46.042 C48.6194967,45.8739992 48.1680013,45.7340006 47.916,45.622 L48.441,43.081 C48.9730027,43.291001 49.5189972,43.4554994 50.079,43.5745 C50.6390028,43.6935006 51.1919973,43.753 51.738,43.753 Z M61.419,40.393 C61.419,41.3590048 61.6359978,42.1359971 62.07,42.724 C62.5040022,43.3120029 63.1479957,43.606 64.002,43.606 C64.2820014,43.606 64.5409988,43.5955001 64.779,43.5745 C65.0170012,43.5534999 65.2129992,43.5290001 65.367,43.501 L65.367,37.81 C65.170999,37.6839994 64.9155016,37.5790004 64.6005,37.495 C64.2854984,37.4109996 63.9670016,37.369 63.645,37.369 C62.1609926,37.369 61.419,38.3769899 61.419,40.393 Z M68.496,45.643 C68.2159986,45.7270004 67.8940018,45.8074996 67.53,45.8845 C67.1659982,45.9615004 66.784502,46.0279997 66.3855,46.084 C65.986498,46.1400003 65.5805021,46.1854998 65.1675,46.2205 C64.7544979,46.2555002 64.3590019,46.273 63.981,46.273 C63.0709954,46.273 62.2590036,46.1400013 61.545,45.874 C60.8309964,45.6079987 60.2290024,45.2265025 59.739,44.7295 C59.2489975,44.2324975 58.8745013,43.6305035 58.6155,42.9235 C58.3564987,42.2164965 58.227,41.4220044 58.227,40.54 C58.227,39.6439955 58.3389989,38.8355036 58.563,38.1145 C58.7870011,37.3934964 59.1089979,36.7810025 59.529,36.277 C59.9490021,35.7729975 60.463497,35.3880013 61.0725,35.122 C61.681503,34.8559987 62.3779961,34.723 63.162,34.723 C63.5960022,34.723 63.9844983,34.7649996 64.3275,34.849 C64.6705017,34.9330004 65.0169982,35.0519992 65.367,35.206 L65.367,30.208 L68.496,29.704 L68.496,45.643 Z" fill="#FFFFFF" fillRule="nonzero"></path>
              </g>
            </a>
            <a href="/docs/reference/configuration/sources/kafka">
              <g id="Source-2" transform="translate(0.000000, 122.000000)">
                <circle id="Oval-Copy-2" fill="#000000" cx="37.5" cy="37.5" r="37.5"></circle>
                <path d="M11.4141414,47 L11.4141414,40.2352092 C11.9393939,40.6661857 12.4646465,41.1625782 12.989899,41.7243867 C13.5151515,42.2861953 14.0221661,42.8710919 14.5109428,43.4790765 C14.9997194,44.0870611 15.4520202,44.6950457 15.8678451,45.3030303 C16.28367,45.9110149 16.6447811,46.4766715 16.9511785,47 L16.9511785,47 L21,47 C19.9640853,45.2299182 18.880752,43.6137566 17.75,42.1515152 C16.619248,40.6892737 15.4483726,39.4271284 14.2373737,38.3650794 C15.3900112,37.1491101 16.4879349,35.9408369 17.5311448,34.7402597 C18.5743547,33.5396825 19.6139169,32.2929293 20.6498316,31 L20.6498316,31 L16.6010101,31 C15.8277217,31.9081289 14.9924242,32.8970659 14.0951178,33.966811 C13.1978114,35.036556 12.3041526,36.0639731 11.4141414,37.049062 L11.4141414,37.049062 L11.4141414,31 L8,31 L8,47 L11.4141414,47 Z M26.9889625,47 C28.1810155,47 29.196468,46.9252874 30.0353201,46.7758621 C30.8741722,46.6264368 31.5290655,46.4966727 32,46.3865699 L32,46.3865699 L32,39.0725953 C32,38.3018754 31.9190581,37.6019359 31.7571744,36.9727768 C31.5952907,36.3436177 31.3193525,35.8088324 30.9293598,35.3684211 C30.5393672,34.9280097 30.0206034,34.5898367 29.3730684,34.353902 C28.7255335,34.1179673 27.9161148,34 26.9448124,34 C26.5621781,34 26.183223,34.0196612 25.807947,34.0589837 C25.4326711,34.0983061 25.075791,34.145493 24.7373068,34.2005445 C24.3988227,34.2555959 24.0860927,34.3185118 23.799117,34.3892922 C23.5121413,34.4600726 23.2729948,34.5269208 23.0816777,34.5898367 L23.0816777,34.5898367 L23.5011038,37.4210526 C23.8984547,37.2637629 24.3693893,37.1418633 24.9139073,37.0553539 C25.4584253,36.9688445 26.0103017,36.9255898 26.5695364,36.9255898 C27.4083885,36.9255898 27.9970567,37.1064731 28.3355408,37.4682396 C28.674025,37.830006 28.8432671,38.3097399 28.8432671,38.907441 L28.8432671,38.907441 L28.8432671,39.2377495 C28.6372333,39.1748336 28.3686534,39.1237145 28.0375276,39.084392 C27.7064018,39.0450696 27.3274467,39.0254083 26.9006623,39.0254083 C26.267844,39.0254083 25.6571008,39.0922565 25.0684327,39.2259528 C24.4797645,39.3596491 23.9573216,39.5759226 23.5011038,39.8747731 C23.0448859,40.1736237 22.6806475,40.5747126 22.4083885,41.0780399 C22.1361295,41.5813672 22,42.1947973 22,42.9183303 C22,43.6733212 22.1177336,44.3103448 22.3532009,44.8294011 C22.5886681,45.3484574 22.9271523,45.7692075 23.3686534,46.0916515 C23.8101545,46.4140956 24.3362767,46.646098 24.9470199,46.7876588 C25.5577631,46.9292196 26.2384106,47 26.9889625,47 Z M28.1728395,44 C27.4979424,44 26.9670782,43.8902439 26.5802469,43.6707317 C26.1934156,43.4512195 26,43.0569106 26,42.4878049 C26,42.195122 26.0658436,41.9512195 26.1975309,41.7560976 C26.3292181,41.5609756 26.5102881,41.4065041 26.7407407,41.2926829 C26.9711934,41.1788618 27.2386831,41.101626 27.5432099,41.0609756 C27.8477366,41.0203252 28.1728395,41 28.5185185,41 C28.7654321,41 29.037037,41.0162602 29.3333333,41.0487805 C29.6296296,41.0813008 29.8518519,41.1138211 30,41.1463415 L30,41.1463415 L30,43.902439 C29.8024691,43.9349593 29.5390947,43.9593496 29.2098765,43.9756098 C28.8806584,43.9918699 28.5349794,44 28.1728395,44 Z M38.735376,47 L38.735376,37.6752577 L43.3481894,37.6752577 L43.3481894,34.7989691 L38.735376,34.7989691 L38.735376,34.1958763 C38.735376,33.871134 38.7729805,33.5695876 38.8481894,33.2912371 C38.9233983,33.0128866 39.0487465,32.7731959 39.224234,32.5721649 C39.3997214,32.371134 39.6420613,32.2126289 39.9512535,32.0966495 C40.2604457,31.9806701 40.6406685,31.9226804 41.091922,31.9226804 C41.4261838,31.9226804 41.7896936,31.9574742 42.1824513,32.0270619 C42.5752089,32.0966495 42.9387187,32.2010309 43.2729805,32.3402062 L43.2729805,32.3402062 L44,29.5798969 C43.6824513,29.4561856 43.2311978,29.3286082 42.6462396,29.1971649 C42.0612813,29.0657216 41.4178273,29 40.7158774,29 C38.8272981,29 37.402507,29.4639175 36.4415042,30.3917526 C35.4805014,31.3195876 35,32.5721649 35,34.1494845 L35,34.1494845 L35,47 L38.735376,47 Z M48.2845691,47 L48.2845691,41.6649485 C48.6519706,41.9742268 49.0230461,42.3530928 49.3977956,42.8015464 C49.7725451,43.25 50.1289245,43.7177835 50.4669339,44.2048969 C50.8049432,44.6920103 51.1245825,45.1791237 51.4258517,45.6662371 C51.7271209,46.1533505 51.995324,46.5979381 52.2304609,47 L52.2304609,47 L56,47 C55.750167,46.4896907 55.4415498,45.9252577 55.0741483,45.306701 C54.7067468,44.6881443 54.3026052,44.0695876 53.8617234,43.4510309 C53.4208417,42.8324742 52.9652639,42.2332474 52.49499,41.6533505 C52.0247161,41.0734536 51.5691383,40.5747423 51.1282565,40.1572165 C52.01002,39.2448454 52.8219773,38.3595361 53.5641283,37.5012887 C54.3062792,36.6430412 55.0668003,35.742268 55.8456914,34.7989691 L55.8456914,34.7989691 L51.9438878,34.7989691 C51.738143,35.0618557 51.491984,35.371134 51.2054108,35.7268041 C50.9188377,36.0824742 50.6138945,36.4574742 50.2905812,36.8518041 C49.9672679,37.246134 49.6329325,37.6443299 49.2875752,38.0463918 C48.9422178,38.4484536 48.6078824,38.8350515 48.2845691,39.2061856 L48.2845691,39.2061856 L48.2845691,29 L45,29.556701 L45,47 L48.2845691,47 Z M61.9889625,47 C63.1810155,47 64.196468,46.9252874 65.0353201,46.7758621 C65.8741722,46.6264368 66.5290655,46.4966727 67,46.3865699 L67,46.3865699 L67,39.0725953 C67,38.3018754 66.9190581,37.6019359 66.7571744,36.9727768 C66.5952907,36.3436177 66.3193525,35.8088324 65.9293598,35.3684211 C65.5393672,34.9280097 65.0206034,34.5898367 64.3730684,34.353902 C63.7255335,34.1179673 62.9161148,34 61.9448124,34 C61.5621781,34 61.183223,34.0196612 60.807947,34.0589837 C60.4326711,34.0983061 60.075791,34.145493 59.7373068,34.2005445 C59.3988227,34.2555959 59.0860927,34.3185118 58.799117,34.3892922 C58.5121413,34.4600726 58.2729948,34.5269208 58.0816777,34.5898367 L58.0816777,34.5898367 L58.5011038,37.4210526 C58.8984547,37.2637629 59.3693893,37.1418633 59.9139073,37.0553539 C60.4584253,36.9688445 61.0103017,36.9255898 61.5695364,36.9255898 C62.4083885,36.9255898 62.9970567,37.1064731 63.3355408,37.4682396 C63.674025,37.830006 63.8432671,38.3097399 63.8432671,38.907441 L63.8432671,38.907441 L63.8432671,39.2377495 C63.6372333,39.1748336 63.3686534,39.1237145 63.0375276,39.084392 C62.7064018,39.0450696 62.3274467,39.0254083 61.9006623,39.0254083 C61.267844,39.0254083 60.6571008,39.0922565 60.0684327,39.2259528 C59.4797645,39.3596491 58.9573216,39.5759226 58.5011038,39.8747731 C58.0448859,40.1736237 57.6806475,40.5747126 57.4083885,41.0780399 C57.1361295,41.5813672 57,42.1947973 57,42.9183303 C57,43.6733212 57.1177336,44.3103448 57.3532009,44.8294011 C57.5886681,45.3484574 57.9271523,45.7692075 58.3686534,46.0916515 C58.8101545,46.4140956 59.3362767,46.646098 59.9470199,46.7876588 C60.5577631,46.9292196 61.2384106,47 61.9889625,47 Z M62.1728395,44 C61.4979424,44 60.9670782,43.8902439 60.5802469,43.6707317 C60.1934156,43.4512195 60,43.0569106 60,42.4878049 C60,42.195122 60.0658436,41.9512195 60.1975309,41.7560976 C60.3292181,41.5609756 60.5102881,41.4065041 60.7407407,41.2926829 C60.9711934,41.1788618 61.2386831,41.101626 61.5432099,41.0609756 C61.8477366,41.0203252 62.1728395,41 62.5185185,41 C62.7654321,41 63.037037,41.0162602 63.3333333,41.0487805 C63.6296296,41.0813008 63.8518519,41.1138211 64,41.1463415 L64,41.1463415 L64,43.902439 C63.8024691,43.9349593 63.5390947,43.9593496 63.2098765,43.9756098 C62.8806584,43.9918699 62.5349794,44 62.1728395,44 Z" id="Kafka" fill="#FFFFFF"></path>
              </g>
            </a>
            <a href="/docs/reference/configuration/sources/file">
              <g id="Source-1" transform="translate(82.000000, 56.000000)">
                <circle id="Oval-Copy-2" fill="#000000" cx="37.5" cy="37.5" r="37.5"></circle>
                <path d="M13.455,49 L13.455,28.21 L27.375,28.21 L27.375,32.14 L18.135,32.14 L18.135,36.49 L26.355,36.49 L26.355,40.42 L18.135,40.42 L18.135,49 L13.455,49 Z M34.845,49 L30.375,49 L30.375,33.22 L34.845,33.22 L34.845,49 Z M35.265,28.66 C35.265,29.4800041 35.0000026,30.1249976 34.47,30.595 C33.9399973,31.0650023 33.3150036,31.3 32.595,31.3 C31.8749964,31.3 31.2500026,31.0650023 30.72,30.595 C30.1899973,30.1249976 29.925,29.4800041 29.925,28.66 C29.925,27.8399959 30.1899973,27.1950023 30.72,26.725 C31.2500026,26.2549976 31.8749964,26.02 32.595,26.02 C33.3150036,26.02 33.9399973,26.2549976 34.47,26.725 C35.0000026,27.1950023 35.265,27.8399959 35.265,28.66 Z M45.345,49.3 C44.045,49.2799999 42.9900041,49.1400013 42.18,48.88 C41.369996,48.6199987 40.7300023,48.2550023 40.26,47.785 C39.7899976,47.3149976 39.4700008,46.7450033 39.3,46.075 C39.1299992,45.4049966 39.045,44.6500042 39.045,43.81 L39.045,26.44 L43.515,25.72 L43.515,42.91 C43.515,43.310002 43.545,43.6699984 43.605,43.99 C43.6650003,44.3100016 43.7799992,44.5799989 43.95,44.8 C44.1200008,45.0200011 44.3649984,45.1999993 44.685,45.34 C45.0050016,45.4800007 45.4349973,45.5699998 45.975,45.61 L45.345,49.3 Z M47.775,41.23 C47.775,39.829993 47.9899979,38.6050052 48.42,37.555 C48.8500022,36.5049947 49.4149965,35.6300035 50.115,34.93 C50.8150035,34.2299965 51.6199955,33.7000018 52.53,33.34 C53.4400046,32.9799982 54.3749952,32.8 55.335,32.8 C57.5750112,32.8 59.3449935,33.4849931 60.645,34.855 C61.9450065,36.2250068 62.595,38.2399867 62.595,40.9 C62.595,41.1600013 62.5850001,41.4449984 62.565,41.755 C62.5449999,42.0650015 62.5250001,42.3399988 62.505,42.58 L52.365,42.58 C52.4650005,43.5000046 52.8949962,44.2299973 53.655,44.77 C54.4150038,45.3100027 55.4349936,45.58 56.715,45.58 C57.5350041,45.58 58.3399961,45.5050007 59.13,45.355 C59.920004,45.2049992 60.5649975,45.0200011 61.065,44.8 L61.665,48.43 C61.4249988,48.5500006 61.105002,48.6699994 60.705,48.79 C60.304998,48.9100006 59.8600025,49.0149995 59.37,49.105 C58.8799976,49.1950004 58.3550028,49.2699997 57.795,49.33 C57.2349972,49.3900003 56.6750028,49.42 56.115,49.42 C54.6949929,49.42 53.4600053,49.2100021 52.41,48.79 C51.3599948,48.3699979 50.4900035,47.7950036 49.8,47.065 C49.1099966,46.3349963 48.6000017,45.470005 48.27,44.47 C47.9399984,43.469995 47.775,42.3900058 47.775,41.23 Z M58.275,39.52 C58.2549999,39.1399981 58.1900006,38.7700018 58.08,38.41 C57.9699994,38.0499982 57.8000011,37.7300014 57.57,37.45 C57.3399988,37.1699986 57.0500017,36.9400009 56.7,36.76 C56.3499982,36.5799991 55.9150026,36.49 55.395,36.49 C54.8949975,36.49 54.4650018,36.5749991 54.105,36.745 C53.7449982,36.9150008 53.4450012,37.1399986 53.205,37.42 C52.9649988,37.7000014 52.7800006,38.0249982 52.65,38.395 C52.5199993,38.7650019 52.4250003,39.1399981 52.365,39.52 L58.275,39.52 Z" fill="#FFFFFF" fillRule="nonzero"></path>
              </g>
            </a>
            <a href="/docs/reference/configuration/sources">
              <path className="fade-in" d="M73.344,13.382 C73.8280024,13.382 74.2276651,13.3416671 74.543,13.261 C74.8583349,13.1803329 75.1113324,13.070334 75.302,12.931 C75.4926676,12.791666 75.6246663,12.6266676 75.698,12.436 C75.7713337,12.2453324 75.808,12.0326678 75.808,11.798 C75.808,11.2993308 75.5733357,10.8850016 75.104,10.555 C74.6346643,10.2249983 73.8280057,9.86933524 72.684,9.488 C72.1853308,9.31199912 71.6866692,9.11033447 71.188,8.883 C70.6893308,8.65566553 70.242002,8.36966839 69.846,8.025 C69.449998,7.68033161 69.1273346,7.26233579 68.878,6.771 C68.6286654,6.27966421 68.504,5.68200352 68.504,4.978 C68.504,4.27399648 68.6359987,3.63966949 68.9,3.075 C69.1640013,2.51033051 69.5379976,2.03000198 70.022,1.634 C70.5060024,1.23799802 71.0926632,0.93366773 71.782,0.721 C72.4713368,0.50833227 73.2486623,0.402 74.114,0.402 C75.1406718,0.402 76.0279963,0.5119989 76.776,0.732 C77.5240037,0.9520011 78.1399976,1.19399868 78.624,1.458 L77.634,4.164 C77.2086645,3.9439989 76.7356693,3.74966751 76.215,3.581 C75.6943307,3.41233249 75.067337,3.328 74.334,3.328 C73.5126626,3.328 72.9223351,3.44166553 72.563,3.669 C72.2036649,3.89633447 72.024,4.24466432 72.024,4.714 C72.024,4.99266806 72.0899993,5.22733238 72.222,5.418 C72.3540007,5.60866762 72.5409988,5.78099923 72.783,5.935 C73.0250012,6.08900077 73.3036651,6.22833271 73.619,6.353 C73.9343349,6.47766729 74.2826648,6.60599934 74.664,6.738 C75.456004,7.0313348 76.1453304,7.32099857 76.732,7.607 C77.3186696,7.89300143 77.8063314,8.22666476 78.195,8.608 C78.5836686,8.98933524 78.8733324,9.4366641 79.064,9.95 C79.2546676,10.4633359 79.35,11.086663 79.35,11.82 C79.35,13.2426738 78.8513383,14.3463294 77.854,15.131 C76.8566617,15.9156706 75.3533434,16.308 73.344,16.308 C72.66933,16.308 72.0606694,16.2676671 71.518,16.187 C70.9753306,16.1063329 70.4950021,16.0073339 70.077,15.89 C69.6589979,15.7726661 69.2996682,15.6480007 68.999,15.516 C68.6983318,15.3839993 68.4453344,15.2593339 68.24,15.142 L69.208,12.414 C69.6626689,12.6633346 70.2236633,12.886999 70.891,13.085 C71.5583367,13.283001 72.3759952,13.382 73.344,13.382 Z M92.594,10.192 C92.594,11.1013379 92.4620013,11.9336629 92.198,12.689 C91.9339987,13.4443371 91.5526692,14.089664 91.054,14.625 C90.5553308,15.160336 89.9576702,15.5746652 89.261,15.868 C88.5643299,16.1613348 87.7833377,16.308 86.918,16.308 C86.0673291,16.308 85.2936701,16.1613348 84.597,15.868 C83.9003298,15.5746652 83.3026692,15.160336 82.804,14.625 C82.3053308,14.089664 81.9166681,13.4443371 81.638,12.689 C81.3593319,11.9336629 81.22,11.1013379 81.22,10.192 C81.22,9.28266212 81.3629986,8.45400374 81.649,7.706 C81.9350014,6.95799626 82.3309975,6.32000264 82.837,5.792 C83.3430025,5.26399736 83.9443299,4.8533348 84.641,4.56 C85.3376702,4.2666652 86.0966626,4.12 86.918,4.12 C87.7540042,4.12 88.5203298,4.2666652 89.217,4.56 C89.9136702,4.8533348 90.5113308,5.26399736 91.01,5.792 C91.5086692,6.32000264 91.8973319,6.95799626 92.176,7.706 C92.4546681,8.45400374 92.594,9.28266212 92.594,10.192 Z M89.25,10.192 C89.25,9.17999494 89.0483354,8.38433623 88.645,7.805 C88.2416646,7.22566377 87.6660037,6.936 86.918,6.936 C86.1699963,6.936 85.5906687,7.22566377 85.18,7.805 C84.7693313,8.38433623 84.564,9.17999494 84.564,10.192 C84.564,11.2040051 84.7693313,12.006997 85.18,12.601 C85.5906687,13.195003 86.1699963,13.492 86.918,13.492 C87.6660037,13.492 88.2416646,13.195003 88.645,12.601 C89.0483354,12.006997 89.25,11.2040051 89.25,10.192 Z M105.002,15.604 C104.444664,15.7653341 103.726004,15.915666 102.846,16.055 C101.965996,16.194334 101.042005,16.264 100.074,16.264 C99.0913284,16.264 98.2736699,16.1320013 97.621,15.868 C96.9683301,15.6039987 96.4513352,15.233669 96.07,14.757 C95.6886648,14.2803309 95.4173341,13.7120033 95.256,13.052 C95.0946659,12.3919967 95.014,11.666004 95.014,10.874 L95.014,4.428 L98.292,4.428 L98.292,10.478 C98.292,11.5340053 98.4313319,12.2966643 98.71,12.766 C98.9886681,13.2353357 99.5093295,13.47 100.272,13.47 C100.506668,13.47 100.755999,13.4590001 101.02,13.437 C101.284001,13.4149999 101.518666,13.3893335 101.724,13.36 L101.724,4.428 L105.002,4.428 L105.002,15.604 Z M115.056,7.332 C114.762665,7.2586663 114.418002,7.18166707 114.022,7.101 C113.625998,7.02033293 113.200669,6.98 112.746,6.98 C112.540666,6.98 112.295001,6.99833315 112.009,7.035 C111.722999,7.07166685 111.506667,7.11199978 111.36,7.156 L111.36,16 L108.082,16 L108.082,5.044 C108.66867,4.83866564 109.361663,4.64433425 110.161,4.461 C110.960337,4.27766575 111.851328,4.186 112.834,4.186 C113.010001,4.186 113.222665,4.19699989 113.472,4.219 C113.721335,4.24100011 113.970665,4.27033315 114.22,4.307 C114.469335,4.34366685 114.718665,4.38766641 114.968,4.439 C115.217335,4.49033359 115.429999,4.5526663 115.606,4.626 L115.056,7.332 Z M116.596,10.214 C116.596,9.37799582 116.731665,8.58967037 117.003,7.849 C117.274335,7.10832963 117.666664,6.46300275 118.18,5.913 C118.693336,5.36299725 119.316663,4.92666828 120.05,4.604 C120.783337,4.28133172 121.619329,4.12 122.558,4.12 C123.174003,4.12 123.738664,4.17499945 124.252,4.285 C124.765336,4.39500055 125.263998,4.55266564 125.748,4.758 L125.066,7.376 C124.757998,7.25866608 124.420669,7.15600044 124.054,7.068 C123.687331,6.97999956 123.276669,6.936 122.822,6.936 C121.853995,6.936 121.131669,7.23666366 120.655,7.838 C120.178331,8.43933634 119.94,9.23132842 119.94,10.214 C119.94,11.2553385 120.163664,12.0619971 120.611,12.634 C121.058336,13.2060029 121.839328,13.492 122.954,13.492 C123.350002,13.492 123.775331,13.4553337 124.23,13.382 C124.684669,13.3086663 125.102665,13.1913341 125.484,13.03 L125.946,15.714 C125.564665,15.8753341 125.088003,16.0146661 124.516,16.132 C123.943997,16.2493339 123.313337,16.308 122.624,16.308 C121.567995,16.308 120.65867,16.1503349 119.896,15.835 C119.13333,15.5196651 118.506336,15.0906694 118.015,14.548 C117.523664,14.0053306 117.164334,13.3636704 116.937,12.623 C116.709666,11.8823296 116.596,11.0793377 116.596,10.214 Z M127.046,10.302 C127.046,9.2753282 127.203665,8.37700385 127.519,7.607 C127.834335,6.83699615 128.248664,6.1953359 128.762,5.682 C129.275336,5.1686641 129.865663,4.78000132 130.533,4.516 C131.200337,4.25199868 131.885996,4.12 132.59,4.12 C134.232675,4.12 135.530662,4.62232831 136.484,5.627 C137.437338,6.63167169 137.914,8.10932358 137.914,10.06 C137.914,10.2506676 137.906667,10.4596655 137.892,10.687 C137.877333,10.9143345 137.862667,11.1159991 137.848,11.292 L130.412,11.292 C130.485334,11.96667 130.800664,12.501998 131.358,12.898 C131.915336,13.294002 132.663329,13.492 133.602,13.492 C134.203336,13.492 134.793664,13.4370005 135.373,13.327 C135.952336,13.2169994 136.425331,13.0813341 136.792,12.92 L137.232,15.582 C137.055999,15.6700004 136.821335,15.7579996 136.528,15.846 C136.234665,15.9340004 135.908335,16.0109997 135.549,16.077 C135.189665,16.1430003 134.804669,16.1979998 134.394,16.242 C133.983331,16.2860002 133.572669,16.308 133.162,16.308 C132.120661,16.308 131.215004,16.1540015 130.445,15.846 C129.674996,15.5379985 129.037003,15.116336 128.531,14.581 C128.024997,14.045664 127.651001,13.411337 127.409,12.678 C127.166999,11.944663 127.046,11.1526709 127.046,10.302 Z M134.746,9.048 C134.731333,8.76933194 134.683667,8.49800132 134.603,8.234 C134.522333,7.96999868 134.397668,7.73533436 134.229,7.53 C134.060332,7.32466564 133.847668,7.15600066 133.591,7.024 C133.334332,6.89199934 133.015335,6.826 132.634,6.826 C132.267332,6.826 131.952001,6.88833271 131.688,7.013 C131.423999,7.13766729 131.204001,7.30266564 131.028,7.508 C130.851999,7.71333436 130.716334,7.95166531 130.621,8.223 C130.525666,8.49433469 130.456,8.76933194 130.412,9.048 L134.746,9.048 Z M143.766,13.646 C144.367336,13.646 144.792665,13.5873339 145.042,13.47 C145.291335,13.3526661 145.416,13.125335 145.416,12.788 C145.416,12.5239987 145.254668,12.293001 144.932,12.095 C144.609332,11.896999 144.118003,11.6733346 143.458,11.424 C142.944664,11.2333324 142.479002,11.0353344 142.061,10.83 C141.642998,10.6246656 141.287335,10.3790014 140.994,10.093 C140.700665,9.80699857 140.473334,9.46600198 140.312,9.07 C140.150666,8.67399802 140.07,8.19733612 140.07,7.64 C140.07,6.55466124 140.473329,5.69666982 141.28,5.066 C142.086671,4.43533018 143.193993,4.12 144.602,4.12 C145.306004,4.12 145.980663,4.18233271 146.626,4.307 C147.271337,4.43166729 147.784665,4.5673326 148.166,4.714 L147.594,7.266 C147.212665,7.13399934 146.798336,7.01666718 146.351,6.914 C145.903664,6.81133282 145.401336,6.76 144.844,6.76 C143.817328,6.76 143.304,7.04599714 143.304,7.618 C143.304,7.75000066 143.326,7.86733282 143.37,7.97 C143.414,8.07266718 143.501999,8.17166619 143.634,8.267 C143.766001,8.36233381 143.945666,8.46499945 144.173,8.575 C144.400334,8.68500055 144.689998,8.80599934 145.042,8.938 C145.76067,9.20200132 146.354664,9.46233205 146.824,9.719 C147.293336,9.97566795 147.663665,10.2543318 147.935,10.555 C148.206335,10.8556682 148.396999,11.1893315 148.507,11.556 C148.617001,11.9226685 148.672,12.3479976 148.672,12.832 C148.672,13.9760057 148.243004,14.8413304 147.385,15.428 C146.526996,16.0146696 145.313341,16.308 143.744,16.308 C142.717328,16.308 141.863003,16.2200009 141.181,16.044 C140.498997,15.8679991 140.026001,15.7213339 139.762,15.604 L140.312,12.942 C140.869336,13.1620011 141.44133,13.3343327 142.028,13.459 C142.61467,13.5836673 143.193997,13.646 143.766,13.646 Z" fill="#000000" fillRule="nonzero"></path>
            </a>
          </g>
          <g id="Transforms" transform="translate(223.721371, 5.000000)">
            <a href="/docs/reference/configuration/transforms/log_to_metric">
              <g id="Aggregate" transform="translate(29.278629, 194.000000)">
                <polygon id="Fill-74" fill="#F44AF5" points="1.13686838e-13 52.3412682 99.9238736 72.9440335 199.84544 52.3412682 99.9238736 31.7363751"></polygon>
                <polygon id="Fill-79" fill="#F44AF5" points="0.206736663 52.3379226 1.13686838e-13 21.1988549 99.7120786 0.579365079 199.84544 21.4061125 199.516959 52.2652672 99.7120786 72.9440335"></polygon>
                <polygon id="Fill-75" fill="#F44AF5" points="5.68434189e-14 21.6857267 5.68434189e-14 52.5454054 97.3605992 72.9440335 97.3605992 42.3477482"></polygon>
                <polygon id="Fill-76" fill="#F44AF5" points="100.776761 42.1888714 199.84544 21.6857267 199.36429 52.4408483 100.776761 72.9440335"></polygon>
                <polygon id="Fill-77" fill="#F44AF5" points="100.776761 0.579365079 199.84544 21.2160983 199.364396 51.8376719 100.981171 31.4242294"></polygon>
                <polygon id="Fill-78" fill="#F44AF5" points="1.13686838e-13 20.9515833 1.13686838e-13 51.8376719 99.0686799 31.3963029 98.8616119 0.579365079"></polygon>
                <polygon id="Fill-74-Copy" fillOpacity="0.33241368" fill="#FFFFFF" points="1.70530257e-13 21.0555539 99.9238736 41.6583192 199.84544 21.0555539 99.9238736 0.450660813"></polygon>
              </g>
            </a>
            <a href="/docs/reference/configuration/transforms/sample">
              <g id="Sample" transform="translate(29.278629, 134.000000)">
                <polygon id="Fill-74" fill="#3084FF" points="1.13686838e-13 52.3412682 99.9238736 72.9440335 199.84544 52.3412682 99.9238736 31.7363751"></polygon>
                <polygon id="Fill-79" fill="#3084FF" points="0.206736663 52.3379226 1.13686838e-13 21.1988549 99.7120786 0.579365079 199.84544 21.4061125 199.516959 52.2652672 99.7120786 72.9440335"></polygon>
                <polygon id="Fill-75" fill="#3084FF" points="5.68434189e-14 21.6857267 5.68434189e-14 52.5454054 97.3605992 72.9440335 97.3605992 42.3477482"></polygon>
                <polygon id="Fill-76" fill="#3084FF" points="100.776761 41.1888714 199.84544 20.6857267 199.36429 51.4408483 100.776761 71.9440335"></polygon>
                <polygon id="Fill-77" fill="#3084FF" points="100.776761 0.579365079 199.84544 21.2160983 199.364396 51.8376719 100.981171 31.4242294"></polygon>
                <polygon id="Fill-78" fill="#3084FF" points="1.13686838e-13 20.9515833 1.13686838e-13 51.8376719 99.0686799 31.3963029 98.8616119 0.579365079"></polygon>
                <polygon id="Fill-74-Copy" fillOpacity="0.33241368" fill="#FFFFFF" points="1.70530257e-13 21.0555539 99.9238736 41.6583192 199.84544 21.0555539 99.9238736 0.450660813"></polygon>
              </g>
            </a>
            <a href="/docs/reference/vrl">
              <g id="Structure" transform="translate(29.278629, 74.000000)">
                <polygon id="Fill-74" fill="#10E7FF" points="1.13686838e-13 52.4612587 99.9238736 72.9443586 199.84544 52.4612587 99.9238736 31.9760433"></polygon>
                <polygon id="Fill-79" fill="#10E7FF" points="0.206736663 52.4579325 1.13686838e-13 21.4997273 99.7120786 1 199.84544 21.7057811 199.516959 52.3856991 99.7120786 72.9443586"></polygon>
                <polygon id="Fill-75" fill="#10E7FF" points="5.68434189e-14 21.9837713 5.68434189e-14 52.6642102 97.3605992 72.9443586 97.3605992 42.5257832"></polygon>
                <polygon id="Fill-76" fill="#10E7FF" points="100.776761 42.3678292 199.84544 21.9837713 199.36429 52.5602604 100.776761 72.9443586"></polygon>
                <polygon id="Fill-77" fill="#10E7FF" points="100.776761 1 199.84544 21.5168705 199.364396 51.9605873 100.981171 31.6657106"></polygon>
                <polygon id="Fill-78" fill="#10E7FF" points="1.13686838e-13 21.2538919 1.13686838e-13 51.9605873 99.0686799 31.6379463 98.8616119 1"></polygon>
                <polygon id="Fill-74-Copy" fillOpacity="0.33241368" fill="#FFFFFF" points="1.70530257e-13 21.3572587 99.9238736 41.8403586 199.84544 21.3572587 99.9238736 0.872043277"></polygon>
              </g>
            </a>
            <path className="fade-in" d="M137.912913,124.27351 L140.194163,124.27351 C139.922479,125.187572 140.600085,125.77351 141.92821,125.77351 C143.154772,125.77351 144.269485,125.17976 144.62449,124.328197 C144.924126,123.609447 144.534243,123.195385 143.275193,122.89851 L141.93672,122.58601 C140.061163,122.156322 139.499548,121.08601 140.21607,119.36726 C141.101952,117.24226 143.395877,115.843822 145.99744,115.843822 C148.481815,115.843822 149.659896,117.234447 148.893181,119.24226 L146.674431,119.24226 C146.936344,118.351635 146.355745,117.757885 145.21512,117.757885 C144.035433,117.757885 143.026198,118.30476 142.667937,119.164135 C142.378072,119.859447 142.751031,120.257885 143.954095,120.539135 L145.192964,120.828197 C147.259917,121.30476 147.829365,122.281322 147.099815,124.031322 C146.152052,126.30476 143.830133,127.695385 141.002008,127.695385 C138.314508,127.695385 137.12274,126.375072 137.912913,124.27351 Z M152.786957,116.74226 L155.060394,116.74226 L154.171256,118.875072 L155.960318,118.875072 L155.214484,120.664135 L153.425422,120.664135 L151.80999,124.539135 C151.497326,125.289135 151.697772,125.632885 152.471209,125.632885 C152.768084,125.632885 152.927591,125.625072 153.156112,125.601635 L152.433076,127.33601 C152.129028,127.390697 151.687611,127.437572 151.250111,127.437572 C149.156361,127.437572 148.611588,126.757885 149.311825,125.078197 L151.151984,120.664135 L149.808234,120.664135 L150.554068,118.875072 L151.897818,118.875072 L152.786957,116.74226 Z M154.294389,127.406322 L157.886769,118.789135 L160.089894,118.789135 L159.516677,120.164135 L159.657302,120.164135 C160.242125,119.30476 161.406331,118.64851 162.500081,118.64851 C162.812581,118.64851 163.179109,118.687572 163.34511,118.757885 L162.482026,120.828197 C162.331651,120.757885 161.829713,120.687572 161.462526,120.687572 C160.220338,120.687572 159.062626,121.421947 158.557803,122.632885 L156.567826,127.406322 L154.294389,127.406322 Z M172.644582,118.789135 L169.052201,127.406322 L166.849076,127.406322 L167.425551,126.02351 L167.284926,126.02351 C166.46179,127.02351 165.419348,127.593822 164.11466,127.593822 C162.255285,127.593822 161.594087,126.406322 162.427858,124.406322 L164.769582,118.789135 L167.043019,118.789135 L164.945564,123.820385 C164.450513,125.007885 164.735536,125.61726 165.782411,125.61726 C166.915223,125.61726 167.820978,124.906322 168.299745,123.757885 L170.371144,118.789135 L172.644582,118.789135 Z M181.283891,121.828197 L179.158891,121.828197 C179.344617,121.007885 178.983407,120.468822 178.014657,120.468822 C176.788095,120.468822 175.719535,121.382885 175.012784,123.078197 C174.296262,124.796947 174.604001,125.726635 175.822751,125.726635 C176.752439,125.726635 177.550715,125.27351 178.077594,124.421947 L180.202594,124.421947 C179.210635,126.351635 177.286535,127.593822 175.044348,127.593822 C172.356848,127.593822 171.510849,125.968822 172.715909,123.078197 C173.907941,120.218822 176.097748,118.601635 178.79306,118.601635 C181.07431,118.601635 181.913877,119.86726 181.283891,121.828197 Z M185.904144,116.74226 L188.177582,116.74226 L187.288443,118.875072 L189.077506,118.875072 L188.331672,120.664135 L186.542609,120.664135 L184.927178,124.539135 C184.614514,125.289135 184.814959,125.632885 185.588397,125.632885 C185.885272,125.632885 186.044779,125.625072 186.273299,125.601635 L185.550264,127.33601 C185.246215,127.390697 184.804799,127.437572 184.367299,127.437572 C182.273549,127.437572 181.728776,126.757885 182.429013,125.078197 L184.269172,120.664135 L182.925422,120.664135 L183.671256,118.875072 L185.015006,118.875072 L185.904144,116.74226 Z M198.800832,118.789135 L195.208451,127.406322 L193.005326,127.406322 L193.581801,126.02351 L193.441176,126.02351 C192.61804,127.02351 191.575598,127.593822 190.27091,127.593822 C188.411535,127.593822 187.750337,126.406322 188.584108,124.406322 L190.925832,118.789135 L193.199269,118.789135 L191.101814,123.820385 C190.606763,125.007885 190.891786,125.61726 191.938661,125.61726 C193.071473,125.61726 193.977228,124.906322 194.455995,123.757885 L196.527394,118.789135 L198.800832,118.789135 Z M197.489701,127.406322 L201.082082,118.789135 L203.285207,118.789135 L202.711989,120.164135 L202.852614,120.164135 C203.437438,119.30476 204.601644,118.64851 205.695394,118.64851 C206.007894,118.64851 206.374422,118.687572 206.540422,118.757885 L205.677339,120.828197 C205.526964,120.757885 205.025026,120.687572 204.657838,120.687572 C203.415651,120.687572 202.257938,121.421947 201.753116,122.632885 L199.763139,127.406322 L197.489701,127.406322 Z M210.926143,120.343822 C209.871455,120.343822 208.823117,121.078197 208.279253,122.195385 L211.951128,122.195385 C212.379763,121.05476 211.988643,120.343822 210.926143,120.343822 Z M210.848311,124.953197 L212.965499,124.953197 C211.955379,126.570385 210.036535,127.593822 207.92716,127.593822 C205.30216,127.593822 204.445752,125.937572 205.611729,123.140697 C206.780962,120.33601 209.082123,118.601635 211.644623,118.601635 C214.183685,118.601635 215.009503,120.218822 213.869582,122.953197 L213.579716,123.64851 L207.665654,123.64851 L207.6168,123.765697 C207.113916,125.046947 207.53171,125.843822 208.71921,125.843822 C209.617647,125.843822 210.368369,125.52351 210.848311,124.953197 Z" id="Structure-Label" fill="#FFFFFF" fillRule="nonzero" transform="translate(176.022165, 121.769603) rotate(-12.000000) translate(-176.022165, -121.769603) "></path>
            <path className="fade-in" d="M138.906466,192.228116 L141.120295,191.795761 C141.019591,192.646949 141.781626,193.031151 143.0705,192.779438 C144.260812,192.546973 145.236733,191.816245 145.429437,191.003947 C145.592086,190.318337 145.13991,190.029974 143.865147,190.008865 L142.510521,189.989139 C140.613794,189.968679 139.877972,189.138723 140.266916,187.499221 C140.747792,185.4722 142.724623,183.813976 145.249299,183.320913 C147.660251,182.850061 149.051422,183.843417 148.665299,185.745327 L146.512122,186.165837 C146.607523,185.337006 145.938235,184.927584 144.83132,185.143761 C143.686497,185.367342 142.804582,186.037068 142.61011,186.856819 C142.452764,187.520072 142.885731,187.797973 144.103378,187.816022 L145.357164,187.83412 C147.447988,187.859316 148.174698,188.605768 147.778683,190.275079 C147.264216,192.443693 145.258826,194.100388 142.514284,194.63639 C139.90621,195.145739 138.514291,194.216493 138.906466,192.228116 Z M152.216708,191.042957 C153.240225,190.843067 154.165591,190.051143 154.363598,189.216487 L154.5068,188.612852 L152.791351,189.047373 C151.827738,189.299527 151.285703,189.725201 151.146037,190.313931 C151.001067,190.925018 151.450965,191.192505 152.216708,191.042957 Z M151.101698,192.817151 C149.494396,193.131053 148.594567,192.382875 148.951688,190.877514 C149.305273,189.387058 150.670339,188.296051 152.775987,187.771112 L154.828563,187.256537 L154.987676,186.585832 C155.150325,185.900222 154.741871,185.610427 153.832078,185.788107 C153.036009,185.943577 152.441651,186.315506 152.14301,186.842893 L150.111138,187.239712 C150.615846,185.691314 152.259901,184.460537 154.360007,184.050392 C156.573837,183.618037 157.580829,184.402142 157.163598,186.16088 L155.83235,191.772449 L153.694336,192.189997 L153.948917,191.116868 L153.812448,191.14352 C153.201736,191.980599 152.185868,192.605416 151.101698,192.817151 Z M158.00069,191.348978 L159.950712,183.129111 L162.088726,182.711562 L161.770499,184.052973 L161.906968,184.026321 C162.470899,183.020704 163.480683,182.269149 164.580016,182.054452 C165.755165,181.824949 166.424454,182.234371 166.53175,183.123114 L166.668219,183.096462 C167.272336,182.104317 168.378402,181.312638 169.54597,181.084615 C171.213924,180.758868 171.999791,181.621693 171.598472,183.313361 L170.260152,188.954738 L168.053904,189.385613 L169.256093,184.318061 C169.487692,183.341812 169.140129,182.947734 168.207591,183.129856 C167.312961,183.304575 166.588253,184.043098 166.381405,184.915015 L165.188055,189.945305 L163.065205,190.359892 L164.28861,185.202913 C164.495457,184.330996 164.098359,183.932378 163.234055,184.101174 C162.369752,184.26997 161.598296,185.053157 161.380841,185.969788 L160.206938,190.918103 L158.00069,191.348978 Z M179.209496,179.239997 C181.324764,178.826891 182.246419,180.15358 181.61527,182.814045 C180.985888,185.467057 179.328028,187.304626 177.235504,187.71329 C176.03761,187.947236 175.242222,187.612189 175.039662,186.820228 L174.903193,186.84688 L173.957353,190.833851 L171.751105,191.264725 L174.348188,180.317323 L176.486202,179.899775 L176.166207,181.248638 L176.302676,181.221986 C176.976795,180.209063 178.057091,179.465058 179.209496,179.239997 Z M176.876623,186.056375 C178.066935,185.82391 178.987712,184.77703 179.348369,183.256764 C179.70549,181.751403 179.210273,180.973956 178.035124,181.203459 C176.859975,181.432962 175.912917,182.49919 175.555796,184.00455 C175.198675,185.509911 175.709055,186.284398 176.876623,186.056375 Z M182.512032,186.56198 L185.206351,175.204702 L187.412599,174.773828 L184.71828,186.131105 L182.512032,186.56198 Z M192.222559,178.240803 C191.199042,178.440693 190.312604,179.281871 189.983974,180.362354 L193.54733,179.66644 C193.759958,178.587291 193.253658,178.039432 192.222559,178.240803 Z M192.968742,182.288211 L195.023358,181.886951 C194.331388,183.493242 192.651703,184.752298 190.604668,185.152078 C188.057248,185.649583 186.930889,184.362871 187.563807,181.694954 C188.198492,179.019585 190.122457,177.066083 192.609225,176.580424 C195.073248,176.099207 196.162956,177.357542 195.544182,179.96584 L195.386836,180.629093 L189.647558,181.749959 L189.621039,181.861743 C189.361426,183.077994 189.908932,183.695983 191.061337,183.470922 C191.933222,183.300645 192.604654,182.878129 192.968742,182.288211 Z" id="Sample-Label" fill="#FFFFFF" fillRule="nonzero"></path>
            <path className="fade-in" d="M14.379204,161.590782 C10.7929992,155.128919 9,147.926538 9,140 C9,127.153085 13.3775094,116.876139 22.1211537,109.246524 L22.8746297,108.589049 L24.1895806,110.096 L23.4361047,110.753476 C15.1415768,117.991197 11,127.714251 11,140 C11,147.470383 12.6548675,154.239403 15.964406,160.322645 L20.4991214,156.694941 L23.7786292,172 L9.56687739,165.440571 L14.379204,161.590782 L14.379204,161.590782 Z" id="Line" className="fade-in" fill="url(#diagram-gradient-2)" fillRule="nonzero"></path>
            <path className="fade-in" d="M14.3141739,225.644231 C4.76775209,211.063968 0,193.174861 0,172 C0,145.12301 7.34863133,124.202725 22.0670751,109.297369 L22.7697064,108.585815 L24.1928146,109.991077 L23.4901833,110.702631 C9.17045481,125.204209 2,145.617256 2,172 C2,192.72483 6.63088086,210.179546 15.8818472,224.386978 L20.4804629,220.698951 L23.7786292,236 L9.55889027,229.457904 L14.3141739,225.644231 L14.3141739,225.644231 Z" id="Line" className="fade-in" fill="url(#diagram-gradient-3)" fillRule="nonzero"></path>
            <path className="fade-in" d="M145.190506,252.517035 L145.132342,249.999273 L141.407669,250.646141 L139.849604,253.444596 L137.623616,253.831186 L144.419262,242.047176 L147.027268,241.594241 L147.592809,252.099824 L145.190506,252.517035 Z M144.939735,244.161318 L142.388176,248.763668 L145.165151,248.281388 L145.071971,244.138352 L144.939735,244.161318 Z M152.87704,249.374389 C154.037786,249.172801 154.989564,248.118343 155.408358,246.620013 C155.827152,245.121682 155.392601,244.300642 154.239201,244.500954 C153.085801,244.701266 152.153905,245.75962 151.737269,247.250227 C151.316316,248.75628 151.716294,249.575977 152.87704,249.374389 Z M151.414255,254.507798 C149.261732,254.88163 148.114989,254.132837 148.401546,252.657472 L150.517337,252.29002 C150.48074,252.79607 150.990327,253.023553 151.915986,252.862792 C153.062039,252.663756 153.885999,251.99157 154.153682,251.033874 L154.594063,249.458311 L154.461826,249.481276 C153.763277,250.455013 152.777242,251.081863 151.638535,251.279623 C149.618249,251.630489 148.821809,250.203589 149.538508,247.639436 C150.276795,244.99805 152.010996,243.094909 154.068015,242.737664 C155.162643,242.547559 155.935901,242.957051 156.108138,243.8163 L156.240375,243.793334 L156.62247,242.426301 L158.694182,242.066504 L156.321734,250.554468 C155.762622,252.554816 153.860638,254.082932 151.414255,254.507798 Z M162.589107,247.687683 C163.749853,247.486095 164.70163,246.431637 165.120424,244.933307 C165.539219,243.434976 165.104667,242.613935 163.951267,242.814248 C162.797868,243.01456 161.865971,244.072914 161.449336,245.56352 C161.028383,247.069574 161.42836,247.889271 162.589107,247.687683 Z M161.126322,252.821092 C158.973799,253.194923 157.827055,252.446131 158.113613,250.970766 L160.229403,250.603314 C160.192807,251.109364 160.702394,251.336846 161.628052,251.176086 C162.774105,250.97705 163.598065,250.304864 163.865748,249.347168 L164.30613,247.771604 L164.173893,247.79457 C163.475344,248.768307 162.489309,249.395157 161.350602,249.592917 C159.330316,249.943783 158.533875,248.516883 159.250575,245.95273 C159.988862,243.311344 161.723063,241.408203 163.780081,241.050958 C164.874709,240.860853 165.647968,241.270345 165.820204,242.129594 L165.952441,242.106628 L166.334537,240.739594 L168.406248,240.379798 L166.0338,248.867762 C165.474689,250.86811 163.572705,252.396226 161.126322,252.821092 Z M168.170342,248.5261 L170.551425,240.007242 L172.623137,239.647446 L172.243199,241.006756 L172.375436,240.98379 C172.825945,240.097221 173.844779,239.30301 174.873289,239.124387 C175.167149,239.073352 175.516333,239.049451 175.680567,239.087065 L175.108503,241.133753 C174.958962,241.093588 174.478829,241.110837 174.133543,241.170803 C172.965451,241.373667 171.961762,242.238733 171.627158,243.435852 L170.308172,248.15482 L168.170342,248.5261 Z M179.98818,239.83069 C178.996403,240.002933 178.095565,240.850136 177.713401,241.967334 L181.166254,241.367673 C181.43735,240.247719 180.987303,239.657171 179.98818,239.83069 Z M180.4483,244.086361 L182.4392,243.740599 C181.676443,245.394196 179.990469,246.649647 178.006915,246.994133 C175.538493,247.422827 174.541538,246.0381 175.314364,243.273139 C176.08935,240.500456 178.052581,238.528145 180.462231,238.109658 C182.849842,237.694999 183.813511,239.048765 183.057954,241.751938 L182.865827,242.439317 L177.30453,243.405154 L177.272149,243.521004 C176.947503,244.78253 177.432576,245.447828 178.549243,245.253895 C179.39409,245.10717 180.062971,244.689718 180.4483,244.086361 Z M187.706773,243.325468 C188.867519,243.12388 189.819297,242.069422 190.238091,240.571092 C190.656885,239.072761 190.222334,238.25172 189.068934,238.452033 C187.915534,238.652345 186.983638,239.710698 186.567002,241.201305 C186.146049,242.707359 186.546027,243.527056 187.706773,243.325468 Z M186.243988,248.458877 C184.091465,248.832708 182.944722,248.083916 183.231279,246.608551 L185.34707,246.241099 C185.310473,246.747149 185.82006,246.974631 186.745719,246.813871 C187.891772,246.614835 188.715732,245.942649 188.983415,244.984953 L189.423796,243.409389 L189.291559,243.432355 C188.59301,244.406092 187.606975,245.032942 186.468269,245.230702 C184.447982,245.581568 183.651542,244.154668 184.368241,241.590515 C185.106528,238.949129 186.840729,237.045988 188.897748,236.688743 C189.992376,236.498637 190.765634,236.908129 190.937871,237.767378 L191.070108,237.744413 L191.452203,236.377379 L193.523915,236.017583 L191.151467,244.505547 C190.592355,246.505895 188.690371,248.034011 186.243988,248.458877 Z M196.647978,242.09597 C197.639755,241.923727 198.574748,241.129379 198.816526,240.264364 L198.991384,239.638772 L197.322893,240.031419 C196.385153,240.260414 195.839873,240.685793 195.669334,241.295938 C195.492317,241.929252 195.905982,242.224833 196.647978,242.09597 Z M195.469952,243.909868 C193.912495,244.180353 193.098502,243.366423 193.534567,241.806306 C193.966313,240.261635 195.34074,239.170517 197.388219,238.697354 L199.384273,238.233122 L199.578559,237.53802 C199.777163,236.827472 199.404548,236.510065 198.522969,236.66317 C197.751587,236.797137 197.159618,237.164489 196.840825,237.704851 L194.871965,238.046786 C195.45193,236.44698 197.102042,235.219803 199.137021,234.866385 C201.282198,234.49383 202.196464,235.349133 201.687003,237.171844 L200.061477,242.987528 L197.989766,243.347325 L198.300623,242.235162 L198.168386,242.258128 C197.531604,243.110912 196.520501,243.727418 195.469952,243.909868 Z M205.719005,231.974355 L207.856835,231.603076 L207.267501,233.711551 L208.949848,233.419376 L208.455498,235.188023 L206.773151,235.480198 L205.702419,239.310981 C205.495181,240.052423 205.723442,240.336112 206.450745,240.2098 C206.729912,240.161317 206.879001,240.128076 207.091179,240.069182 L206.61194,241.783766 C206.332355,241.883761 205.922692,241.998998 205.511288,242.070447 C203.542427,242.412381 202.951509,241.875693 203.415636,240.215172 L204.635321,235.851478 L203.371724,236.070928 L203.866073,234.302281 L205.129671,234.08283 L205.719005,231.974355 Z M213.767364,233.964219 C212.775587,234.136462 211.874749,234.983665 211.492586,236.100863 L214.945438,235.501202 C215.216534,234.381248 214.766488,233.7907 213.767364,233.964219 Z M214.227484,238.21989 L216.218385,237.874128 C215.455627,239.527725 213.769653,240.783176 211.786099,241.127662 C209.317677,241.556356 208.320722,240.171628 209.093549,237.406668 C209.868534,234.633985 211.831765,232.661674 214.241415,232.243187 C216.629026,231.828528 217.592695,233.182294 216.837139,235.885467 L216.645011,236.572846 L211.083715,237.538683 L211.051334,237.654533 C210.726687,238.916059 211.21176,239.581357 212.328427,239.387424 C213.173274,239.240699 213.842156,238.823247 214.227484,238.21989 Z" id="Aggregate-Label" fill="#FFFFFF" fillRule="nonzero"></path>
            <a href="/docs/reference/configuration/transforms">
              <path className="fade-in" d="M82.7326292,2.754 L82.7326292,5.68 L78.1346292,5.68 L78.1346292,18 L74.7026292,18 L74.7026292,5.68 L70.1046292,5.68 L70.1046292,2.754 L82.7326292,2.754 Z M90.1686292,9.332 C89.8752944,9.2586663 89.5306312,9.18166707 89.1346292,9.101 C88.7386272,9.02033293 88.3132982,8.98 87.8586292,8.98 C87.6532949,8.98 87.4076306,8.99833315 87.1216292,9.035 C86.8356278,9.07166685 86.6192966,9.11199978 86.4726292,9.156 L86.4726292,18 L83.1946292,18 L83.1946292,7.044 C83.7812988,6.83866564 84.4742919,6.64433425 85.2736292,6.461 C86.0729665,6.27766575 86.9639576,6.186 87.9466292,6.186 C88.1226301,6.186 88.3352946,6.19699989 88.5846292,6.219 C88.8339638,6.24100011 89.0832946,6.27033315 89.3326292,6.307 C89.5819638,6.34366685 89.8312946,6.38766641 90.0806292,6.439 C90.3299638,6.49033359 90.5426283,6.5526663 90.7186292,6.626 L90.1686292,9.332 Z M96.6806292,15.712 C97.0032975,15.712 97.3112944,15.7046667 97.6046292,15.69 C97.897964,15.6753333 98.1326283,15.6533335 98.3086292,15.624 L98.3086292,13.138 C98.1766286,13.1086665 97.9786305,13.0793335 97.7146292,13.05 C97.4506279,13.0206665 97.2086303,13.006 96.9886292,13.006 C96.6806277,13.006 96.3909639,13.0243331 96.1196292,13.061 C95.8482945,13.0976668 95.6099636,13.1673328 95.4046292,13.27 C95.1992949,13.3726672 95.0379631,13.5119991 94.9206292,13.688 C94.8032953,13.8640009 94.7446292,14.0839987 94.7446292,14.348 C94.7446292,14.8613359 94.9169608,15.216999 95.2616292,15.415 C95.6062976,15.613001 96.0792929,15.712 96.6806292,15.712 Z M96.4166292,6.12 C97.3846341,6.12 98.1912927,6.2299989 98.8366292,6.45 C99.4819658,6.6700011 99.9989606,6.98533128 100.387629,7.396 C100.776298,7.80666872 101.051295,8.3053304 101.212629,8.892 C101.373963,9.4786696 101.454629,10.1313297 101.454629,10.85 L101.454629,17.67 C100.985294,17.7726672 100.332633,17.893666 99.4966292,18.033 C98.660625,18.172334 97.6486352,18.242 96.4606292,18.242 C95.7126255,18.242 95.0342989,18.1760007 94.4256292,18.044 C93.8169595,17.9119993 93.2926314,17.6956682 92.8526292,17.395 C92.412627,17.0943318 92.0752971,16.7020024 91.8406292,16.218 C91.6059614,15.7339976 91.4886292,15.1400035 91.4886292,14.436 C91.4886292,13.76133 91.6242945,13.1893357 91.8956292,12.72 C92.1669639,12.2506643 92.5299603,11.8766681 92.9846292,11.598 C93.4392982,11.3193319 93.9599596,11.1176673 94.5466292,10.993 C95.1332988,10.8683327 95.7419594,10.806 96.3726292,10.806 C96.7979647,10.806 97.1756276,10.8243331 97.5056292,10.861 C97.8356309,10.8976668 98.1032949,10.945333 98.3086292,11.004 L98.3086292,10.696 C98.3086292,10.1386639 98.1399642,9.69133502 97.8026292,9.354 C97.4652942,9.01666498 96.8786334,8.848 96.0426292,8.848 C95.4852931,8.848 94.9352986,8.88833293 94.3926292,8.969 C93.8499598,9.04966707 93.3806312,9.1633326 92.9846292,9.31 L92.5666292,6.67 C92.7572968,6.61133304 92.9956278,6.54900033 93.2816292,6.483 C93.5676306,6.41699967 93.8792942,6.35833359 94.2166292,6.307 C94.5539642,6.25566641 94.9096273,6.21166685 95.2836292,6.175 C95.6576311,6.13833315 96.035294,6.12 96.4166292,6.12 Z M104.424629,6.824 C104.981965,6.66266586 105.700625,6.51233403 106.580629,6.373 C107.460634,6.23366597 108.384624,6.164 109.352629,6.164 C110.335301,6.164 111.152959,6.29233205 111.805629,6.549 C112.458299,6.80566795 112.975294,7.16866432 113.356629,7.638 C113.737964,8.10733568 114.009295,8.66466344 114.170629,9.31 C114.331963,9.95533656 114.412629,10.673996 114.412629,11.466 L114.412629,18 L111.134629,18 L111.134629,11.862 C111.134629,10.8059947 110.995297,10.0580022 110.716629,9.618 C110.437961,9.1779978 109.9173,8.958 109.154629,8.958 C108.919961,8.958 108.670631,8.96899989 108.406629,8.991 C108.142628,9.01300011 107.907964,9.03866652 107.702629,9.068 L107.702629,18 L104.424629,18 L104.424629,6.824 Z M120.704629,15.646 C121.305966,15.646 121.731295,15.5873339 121.980629,15.47 C122.229964,15.3526661 122.354629,15.125335 122.354629,14.788 C122.354629,14.5239987 122.193297,14.293001 121.870629,14.095 C121.547961,13.896999 121.056633,13.6733346 120.396629,13.424 C119.883293,13.2333324 119.417631,13.0353344 118.999629,12.83 C118.581627,12.6246656 118.225964,12.3790014 117.932629,12.093 C117.639294,11.8069986 117.411963,11.466002 117.250629,11.07 C117.089295,10.673998 117.008629,10.1973361 117.008629,9.64 C117.008629,8.55466124 117.411959,7.69666982 118.218629,7.066 C119.0253,6.43533018 120.132622,6.12 121.540629,6.12 C122.244633,6.12 122.919293,6.18233271 123.564629,6.307 C124.209966,6.43166729 124.723294,6.5673326 125.104629,6.714 L124.532629,9.266 C124.151294,9.13399934 123.736965,9.01666718 123.289629,8.914 C122.842294,8.81133282 122.339965,8.76 121.782629,8.76 C120.755957,8.76 120.242629,9.04599714 120.242629,9.618 C120.242629,9.75000066 120.264629,9.86733282 120.308629,9.97 C120.352629,10.0726672 120.440629,10.1716662 120.572629,10.267 C120.70463,10.3623338 120.884295,10.4649994 121.111629,10.575 C121.338964,10.6850005 121.628627,10.8059993 121.980629,10.938 C122.699299,11.2020013 123.293294,11.462332 123.762629,11.719 C124.231965,11.9756679 124.602295,12.2543318 124.873629,12.555 C125.144964,12.8556682 125.335629,13.1893315 125.445629,13.556 C125.55563,13.9226685 125.610629,14.3479976 125.610629,14.832 C125.610629,15.9760057 125.181634,16.8413304 124.323629,17.428 C123.465625,18.0146696 122.25197,18.308 120.682629,18.308 C119.655957,18.308 118.801633,18.2200009 118.119629,18.044 C117.437626,17.8679991 116.964631,17.7213339 116.700629,17.604 L117.250629,14.942 C117.807965,15.1620011 118.37996,15.3343327 118.966629,15.459 C119.553299,15.5836673 120.132626,15.646 120.704629,15.646 Z M133.068629,0.928 C133.684632,0.928 134.249293,0.99033271 134.762629,1.115 C135.275965,1.23966729 135.671961,1.36066608 135.950629,1.478 L135.312629,4.096 C135.019294,3.96399934 134.700298,3.86500033 134.355629,3.799 C134.010961,3.73299967 133.691964,3.7 133.398629,3.7 C133.002627,3.7 132.668964,3.75499945 132.397629,3.865 C132.126295,3.97500055 131.91363,4.12533238 131.759629,4.316 C131.605628,4.50666762 131.49563,4.73399868 131.429629,4.998 C131.363629,5.26200132 131.330629,5.54799846 131.330629,5.856 L131.330629,6.428 L135.378629,6.428 L135.378629,9.156 L131.330629,9.156 L131.330629,18 L128.052629,18 L128.052629,5.812 C128.052629,4.31599252 128.474292,3.1280044 129.317629,2.248 C130.160967,1.3679956 131.411288,0.928 133.068629,0.928 Z M147.984629,12.192 C147.984629,13.1013379 147.852631,13.9336629 147.588629,14.689 C147.324628,15.4443371 146.943298,16.089664 146.444629,16.625 C145.94596,17.160336 145.348299,17.5746652 144.651629,17.868 C143.954959,18.1613348 143.173967,18.308 142.308629,18.308 C141.457958,18.308 140.684299,18.1613348 139.987629,17.868 C139.290959,17.5746652 138.693298,17.160336 138.194629,16.625 C137.69596,16.089664 137.307297,15.4443371 137.028629,14.689 C136.749961,13.9336629 136.610629,13.1013379 136.610629,12.192 C136.610629,11.2826621 136.753628,10.4540037 137.039629,9.706 C137.325631,8.95799626 137.721627,8.32000264 138.227629,7.792 C138.733632,7.26399736 139.334959,6.8533348 140.031629,6.56 C140.728299,6.2666652 141.487292,6.12 142.308629,6.12 C143.144633,6.12 143.910959,6.2666652 144.607629,6.56 C145.304299,6.8533348 145.90196,7.26399736 146.400629,7.792 C146.899298,8.32000264 147.287961,8.95799626 147.566629,9.706 C147.845297,10.4540037 147.984629,11.2826621 147.984629,12.192 Z M144.640629,12.192 C144.640629,11.1799949 144.438965,10.3843362 144.035629,9.805 C143.632294,9.22566377 143.056633,8.936 142.308629,8.936 C141.560625,8.936 140.981298,9.22566377 140.570629,9.805 C140.15996,10.3843362 139.954629,11.1799949 139.954629,12.192 C139.954629,13.2040051 140.15996,14.006997 140.570629,14.601 C140.981298,15.195003 141.560625,15.492 142.308629,15.492 C143.056633,15.492 143.632294,15.195003 144.035629,14.601 C144.438965,14.006997 144.640629,13.2040051 144.640629,12.192 Z M157.488629,9.332 C157.195294,9.2586663 156.850631,9.18166707 156.454629,9.101 C156.058627,9.02033293 155.633298,8.98 155.178629,8.98 C154.973295,8.98 154.727631,8.99833315 154.441629,9.035 C154.155628,9.07166685 153.939297,9.11199978 153.792629,9.156 L153.792629,18 L150.514629,18 L150.514629,7.044 C151.101299,6.83866564 151.794292,6.64433425 152.593629,6.461 C153.392967,6.27766575 154.283958,6.186 155.266629,6.186 C155.44263,6.186 155.655295,6.19699989 155.904629,6.219 C156.153964,6.24100011 156.403295,6.27033315 156.652629,6.307 C156.901964,6.34366685 157.151295,6.38766641 157.400629,6.439 C157.649964,6.49033359 157.862628,6.5526663 158.038629,6.626 L157.488629,9.332 Z M166.156629,11.862 C166.156629,10.8059947 166.020964,10.0580022 165.749629,9.618 C165.478295,9.1779978 165.012633,8.958 164.352629,8.958 C164.147295,8.958 163.93463,8.96899989 163.714629,8.991 C163.494628,9.01300011 163.281964,9.03866652 163.076629,9.068 L163.076629,18 L159.798629,18 L159.798629,6.824 C160.077297,6.7506663 160.403627,6.67366707 160.777629,6.593 C161.151631,6.51233293 161.547627,6.43900033 161.965629,6.373 C162.383631,6.30699967 162.812627,6.25566685 163.252629,6.219 C163.692631,6.18233315 164.125294,6.164 164.550629,6.164 C165.386633,6.164 166.06496,6.27033227 166.585629,6.483 C167.106298,6.69566773 167.535294,6.9486652 167.872629,7.242 C168.341965,6.90466498 168.88096,6.64066762 169.489629,6.45 C170.098299,6.25933238 170.659293,6.164 171.172629,6.164 C172.096634,6.164 172.855626,6.29233205 173.449629,6.549 C174.043632,6.80566795 174.516627,7.16866432 174.868629,7.638 C175.220631,8.10733568 175.462629,8.66466344 175.594629,9.31 C175.72663,9.95533656 175.792629,10.673996 175.792629,11.466 L175.792629,18 L172.514629,18 L172.514629,11.862 C172.514629,10.8059947 172.378964,10.0580022 172.107629,9.618 C171.836295,9.1779978 171.370633,8.958 170.710629,8.958 C170.534628,8.958 170.288964,9.00199956 169.973629,9.09 C169.658294,9.17800044 169.397964,9.28799934 169.192629,9.42 C169.295296,9.75733502 169.361296,10.1129981 169.390629,10.487 C169.419963,10.8610019 169.434629,11.2606645 169.434629,11.686 L169.434629,18 L166.156629,18 L166.156629,11.862 Z M182.084629,15.646 C182.685966,15.646 183.111295,15.5873339 183.360629,15.47 C183.609964,15.3526661 183.734629,15.125335 183.734629,14.788 C183.734629,14.5239987 183.573297,14.293001 183.250629,14.095 C182.927961,13.896999 182.436633,13.6733346 181.776629,13.424 C181.263293,13.2333324 180.797631,13.0353344 180.379629,12.83 C179.961627,12.6246656 179.605964,12.3790014 179.312629,12.093 C179.019294,11.8069986 178.791963,11.466002 178.630629,11.07 C178.469295,10.673998 178.388629,10.1973361 178.388629,9.64 C178.388629,8.55466124 178.791959,7.69666982 179.598629,7.066 C180.4053,6.43533018 181.512622,6.12 182.920629,6.12 C183.624633,6.12 184.299293,6.18233271 184.944629,6.307 C185.589966,6.43166729 186.103294,6.5673326 186.484629,6.714 L185.912629,9.266 C185.531294,9.13399934 185.116965,9.01666718 184.669629,8.914 C184.222294,8.81133282 183.719965,8.76 183.162629,8.76 C182.135957,8.76 181.622629,9.04599714 181.622629,9.618 C181.622629,9.75000066 181.644629,9.86733282 181.688629,9.97 C181.732629,10.0726672 181.820629,10.1716662 181.952629,10.267 C182.08463,10.3623338 182.264295,10.4649994 182.491629,10.575 C182.718964,10.6850005 183.008627,10.8059993 183.360629,10.938 C184.079299,11.2020013 184.673294,11.462332 185.142629,11.719 C185.611965,11.9756679 185.982295,12.2543318 186.253629,12.555 C186.524964,12.8556682 186.715629,13.1893315 186.825629,13.556 C186.93563,13.9226685 186.990629,14.3479976 186.990629,14.832 C186.990629,15.9760057 186.561634,16.8413304 185.703629,17.428 C184.845625,18.0146696 183.63197,18.308 182.062629,18.308 C181.035957,18.308 180.181633,18.2200009 179.499629,18.044 C178.817626,17.8679991 178.344631,17.7213339 178.080629,17.604 L178.630629,14.942 C179.187965,15.1620011 179.75996,15.3343327 180.346629,15.459 C180.933299,15.5836673 181.512626,15.646 182.084629,15.646 Z" fill="#000000" fillRule="nonzero"></path>
            </a>
          </g>
          <g id="Logo" transform="translate(328.000000, 38.000000)" fill="#10E7FF">
            <polygon id="Rectangle" points="42.5439541 0 50.0495796 4.17258107e-13 32.1747297 29.4788933 28.466169 23.3786335"></polygon>
            <polygon id="Triangle-Copy" transform="translate(17.780772, 14.661871) scale(1, -1) translate(-17.780772, -14.661871) " points="17.7807717 9.09494702e-13 35.5615434 29.3237412 26.6924774 29.3237412 17.9626561 14.98769 9.22860926 29.3237412 0 29.3237412"></polygon>
          </g>
        </g>
      </g>
    </svg>
  );
};

// Place the components in the DOM
ReactDOM.render(<RotatingGlobe size={750} />, document.getElementById("globe"));
ReactDOM.render(<Diagram className="mx-auto" width="100%" />, document.getElementById("diagram"));
